"""Handler for editing talent orbs"""

from typing import Any

from ... import helper, user_input_handler


def edit_all_orbs(save_stats: dict[str, Any], orb_list: list[str]) -> dict[str, Any]:
    """Handler for editing all talent orbs"""

    val = user_input_handler.colored_input(
        "What do you want to set the value of all talent orbs to?:"
    )
    val = helper.check_int_max(val)
    if val is None:
        print("Error please enter a number")
        return save_stats

    for orb in orb_list:
        try:
            orb_id = orb_list.index(orb)
        except ValueError:
            continue
        save_stats["talent_orbs"][orb_id] = val

    helper.colored_text(f"Set all talent orbs to &{val}&")
    return save_stats


def edit_talent_orbs(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing talent orbs"""

    orb_list = get_talent_orbs_types()

    talent_orbs = save_stats["talent_orbs"]
    print("You have:")
    for orb in talent_orbs:
        amount = talent_orbs[orb]
        text = "orbs" if amount != 1 else "orb"
        try:
            helper.colored_text(f"&{amount}& {orb_list[orb]} {text}")
        except IndexError:
            helper.colored_text(f"&{amount}& Unknown {orb} {text}")

    orbs_str = user_input_handler.colored_input(
        "Enter the name of the orb that you want. You can enter multiple orb names separated by &spaces& to edit multiple at once or you can enter &all& to select all talent orbs to edit (e.g &angel a massive red d strong black b resistant&):"
    ).split(" ")
    if orbs_str[0] == "all":
        return edit_all_orbs(save_stats, orb_list)
    length = len(orbs_str) // 3
    orbs_to_set: list[int] = []

    for i in range(length):
        orb_name = " ".join(orbs_str[i * 3 : i * 3 + 3]).lower()
        orb_name = orb_name.replace("angle", "angel").title()
        try:
            orbs_to_set.append(orb_list.index(orb_name))
        except ValueError:
            helper.colored_text(
                f"Error orb &{orb_name}& does not exist or is not recognized"
            )

    for orb_id in orbs_to_set:
        name = orb_list[orb_id]
        val = helper.check_int_max(
            user_input_handler.colored_input(
                f"What do you want to set the value of &{name}& to?:"
            )
        )
        if val is None:
            print("Error please enter a number")
            continue
        talent_orbs[orb_id] = val
    save_stats["talent_orbs"] = talent_orbs

    return save_stats


ATTRIBUTES = [
    "Red",
    "Floating",
    "Black",
    "Metal",
    "Angel",
    "Alien",
    "Zombie",
]
EFFECTS = [
    "Attack",
    "Defense",
    "Strong",
    "Massive",
    "Resistant",
]
GRADES = [
    "D",
    "C",
    "B",
    "A",
    "S",
]


def create_orb_list(
    attributes: list[str], effects: list[str], grades: list[str], incl_metal: bool
) -> list[str]:
    """Create a list of all possible talent orbs"""

    orb_list: list[str] = []
    for attribute in attributes:
        effects_trim = effects

        if attribute == "Metal" and incl_metal:
            effects_trim = [effects[1]]
        if attribute == "Metal" and not incl_metal:
            effects_trim = []

        for effect in effects_trim:
            for grade in grades:
                orb_list.append(f"{attribute} {grade} {effect}")

    return orb_list


def create_aku_orbs(effects: list[str], grades: list[str]) -> list[str]:
    """Create a list of all possible aku orbs"""

    orb_list: list[str] = []
    for effect in effects:
        for grade in grades:
            orb_list.append(f"Aku {grade} {effect}")

    return orb_list


def get_talent_orbs_types() -> list[str]:
    """Get a list of all possible talent orbs"""

    orb_list = create_orb_list(ATTRIBUTES, EFFECTS[0:2], GRADES, True)
    orb_list += create_orb_list(ATTRIBUTES, EFFECTS[2:], GRADES, False)
    orb_list += create_aku_orbs(EFFECTS, GRADES)
    print(orb_list)
    return orb_list



"""Handler for parsing the save file"""

import collections
import datetime
import json
import struct
import traceback
from typing import Any, Optional, Union


from . import helper
from . import updater

address = 0
save_data_g = None


def re_order(data: dict[str, Any]) -> collections.OrderedDict[str, Any]:
    """Move all unknown vals to the bottom of the json"""

    priority: list[str] = json.loads(
        helper.read_file_string(helper.get_file("order.json"))
    )
    ordered_data = collections.OrderedDict(data)

    for item in ordered_data.copy():
        if "unknown" in item:
            ordered_data.move_to_end(item)

    for i in range(len(priority)):
        ordered_data.move_to_end(priority[len(priority) - 1 - i], False)
    return ordered_data


def set_address(val: int):
    """Set the address to a specific value"""

    global address
    address = val


def next_int_len(number: int) -> dict[str, int]:
    """Get the next int of a specified byte length from the save file"""

    if number < 0:
        raise Exception("Invalid number")
    if save_data_g is None:
        raise Exception("Invalid save data")
    if number > len(save_data_g):
        raise Exception("Byte length is greater than the length of the save data")
    val = convert_little(save_data_g[address : address + number])
    data: dict[str, int] = {}
    set_address(address + number)
    data["Value"] = val
    data["Length"] = number
    return data


def generate_empty_len(length: int) -> dict[str, int]:
    """Generate an empty dict with a length and value of 0"""

    data: dict[str, int] = {}
    data["Length"] = length
    data["Value"] = 0
    return data


def next_int(number: int) -> int:
    return next_int_len(number)["Value"]


def skip(number: int):
    """Skip a number of bytes"""

    set_address(address + number)


def convert_little(byte_data: bytes) -> int:
    """Convert a byte array to an int in little endian"""

    return int.from_bytes(byte_data, byteorder="little", signed=False)


def get_time_data_skip(dst_flag: bool) -> dict[str, Any]:
    year = next_int(4)
    year_2 = next_int(4)

    month = next_int(4)
    month_2 = next_int(4)

    day = next_int(4)
    day_2 = next_int(4)

    time_stamp = get_double()

    hour = next_int(4)
    minute = next_int(4)
    second = next_int(4)
    dst = 0
    if dst_flag:
        dst = next_int(1)

    time = datetime.datetime(year, month, day, hour, minute, second)
    return {
        "time": time.isoformat(),
        "time_stamp": time_stamp,
        "dst": dst,
        "duplicate": {"yy": year_2, "mm": month_2, "dd": day_2},
    }


def get_time_data(dst_flag: bool) -> str:
    if dst_flag:
        _ = next_int(1)
    year = next_int(4)
    month = next_int(4)
    day = next_int(4)
    hour = next_int(4)
    minute = next_int(4)
    second = next_int(4)

    time = datetime.datetime(year, month, day, hour, minute, second)
    return time.isoformat()


def get_length_data(
    length_bytes: int = 4, separator: int = 4, length: Union[int, None] = None
) -> list[int]:

    data: list[int] = []
    if length is None:
        length = next_int(length_bytes)
    if save_data_g is None:
        raise Exception("Invalid save data")
    if length > len(save_data_g):
        raise Exception("Length too large")
    for _ in range(length):
        data.append(next_int(separator))
    return data


def get_length_doubles(
    length_bytes: int = 4, length: Union[int, None] = None
) -> list[float]:
    data: list[float] = []
    if length is None:
        length = next_int(length_bytes)
    if save_data_g is None:
        raise Exception("Invalid save data")
    if length > len(save_data_g):
        raise Exception("Length too large")
    for _ in range(length):
        data.append(get_double())
    return data


def get_equip_slots() -> list[list[int]]:
    length = next_int(1)
    data = get_length_data(1, length=length * 10)
    slots: list[list[int]] = []
    for i in range(length):
        start_pos = 10 * i
        end_pos = 10 * (i + 1)
        slots.append(data[start_pos:end_pos])

    data = slots
    return data


def get_main_story_levels() -> dict[str, Any]:
    chapter_progress: list[int] = []
    for _ in range(10):
        chapter_progress.append(next_int(4))
    times_cleared: list[list[int]] = []
    for _ in range(10):
        chapter_times: list[int] = []
        for _ in range(51):
            chapter_times.append(next_int(4))
        times_cleared.append(chapter_times)
    times_cleared_dict = times_cleared
    return {
        "Chapter Progress": chapter_progress,
        "Times Cleared": times_cleared_dict,
    }


def get_treasures() -> list[list[int]]:
    treasures: list[list[int]] = []
    for _ in range(10):
        chapter: list[int] = []
        for _ in range(49):
            chapter.append(next_int(4))
        treasures.append(chapter)
    return treasures


def get_cat_upgrades() -> dict[str, Any]:
    length = next_int(4)
    data = get_length_data(4, 2, length * 2)
    base_levels = data[1::2]
    plus_levels = data[0::2]

    data_dict = {"Base": base_levels, "Plus": plus_levels}
    return data_dict


def get_blue_upgrades() -> dict[str, Any]:
    length = 11
    data = get_length_data(4, 2, length * 2)
    base_levels = data[1::2]
    plus_levels = data[0::2]
    data_dict = {"Base": base_levels, "Plus": plus_levels}
    return data_dict


def get_utf8_string(length: Union[int, None] = None) -> str:
    data = get_length_data(4, 1, length)
    data = bytes(data).decode("utf-8")
    return data


def read_variable_length_int() -> int:
    """
    Read a variable length int from the save file

    Returns:
        int: The value of the variable length int
    """
    i = 0
    for _ in range(4):
        i_3 = i << 7
        read = next_int(1)
        i = i_3 | (read & 127)
        if (read & 128) == 0:
            return i
    return i


def load_bonus_hash() -> tuple[dict[int, int], dict[int, int]]:
    """
    Get the variable data from the save file

    Returns:
        tuple[dict[int, int], dict[int, int]]: The variable data
    """
    length_1 = read_variable_length_int()
    data_1: dict[int, int] = {}
    for _ in range(length_1):
        key = read_variable_length_int()
        val = read_variable_length_int()
        data_1[key] = val

    length_2 = read_variable_length_int()
    data_2: dict[int, int] = {}
    for _ in range(length_2):
        key = read_variable_length_int()
        val = next_int(1)
        data_2[key] = val

    return (data_1, data_2)


def get_event_stages_current() -> dict[str, Any]:
    unknown_val = next_int(1)
    total_sub_chapters = next_int(2) * unknown_val
    stars_per_sub_chapter = next_int(1)
    stages_per_sub_chapter = next_int(1)

    clear_progress = get_length_data(1, 1, total_sub_chapters * stars_per_sub_chapter)
    clear_progress = list(helper.chunks(clear_progress, stars_per_sub_chapter))

    return {
        "Clear": clear_progress,
        "unknown": unknown_val,
        "total": total_sub_chapters,
        "stages": stages_per_sub_chapter,
        "stars": stars_per_sub_chapter,
    }


def get_event_stages(lengths: dict[str, Any]) -> dict[str, Any]:
    total_sub_chapters = lengths["total"]
    stars_per_sub_chapter = lengths["stars"]
    stages_per_sub_chapter = lengths["stages"]

    clear_progress = get_length_data(1, 1, total_sub_chapters * stars_per_sub_chapter)
    clear_amount = get_length_data(
        1, 2, total_sub_chapters * stages_per_sub_chapter * stars_per_sub_chapter
    )
    unlock_next = get_length_data(1, 1, total_sub_chapters * stars_per_sub_chapter)

    clear_progress = list(helper.chunks(clear_progress, stars_per_sub_chapter))
    clear_amount = list(
        helper.chunks(clear_amount, stages_per_sub_chapter * stars_per_sub_chapter)
    )
    unlock_next = list(helper.chunks(unlock_next, stars_per_sub_chapter))

    clear_amount_sep: list[list[list[int]]] = []

    for clear_amount_val in clear_amount:
        sub_chapter_clears: list[list[int]] = []
        for j in range(stars_per_sub_chapter):
            sub_chapter_clears.append(clear_amount_val[j::stars_per_sub_chapter])
        clear_amount_sep.append(sub_chapter_clears)

    clear_amount = clear_amount_sep
    return {
        "Value": {
            "clear_progress": clear_progress,
            "clear_amount": clear_amount,
            "unlock_next": unlock_next,
        },
        "Lengths": lengths,
    }


def get_purchase_receipts() -> list[dict[str, Any]]:
    total_strs = next_int(4)
    data: list[dict[Any, Any]] = []

    for _ in range(total_strs):
        data_dict: dict[str, Any] = {}
        data_dict["unknown_4"] = next_int(4)

        strings = next_int(4)
        item_packs: list[Any] = []
        for _ in range(strings):
            strings_dict = {}

            strings_dict["Value"] = get_utf8_string()
            strings_dict["unknown_1"] = next_int(1)
            item_packs.append(strings_dict)
        data_dict["item_packs"] = item_packs
        data.append(data_dict)
    return data


def get_dojo_data_maybe() -> dict[int, Any]:
    # everything here is speculative and might not be correct
    dojo_data: dict[int, Any] = {}
    total_subchapters = next_int(4)
    for _ in range(total_subchapters):
        subchapter_id = next_int(4)
        subchapter_data = {}

        total_stages = next_int(4)
        for _ in range(total_stages):
            stage_id = next_int(4)

            score = next_int(4)
            subchapter_data[stage_id] = score
        dojo_data[subchapter_id] = subchapter_data
    return dojo_data


def get_data_before_outbreaks() -> list[dict[str, Any]]:
    data: list[dict[str, Any]] = []

    length = next_int_len(4)
    data.append(length)
    for _ in range(length["Value"]):
        length_2 = next_int_len(4)
        data.append(length_2)

        length_3 = next_int_len(4)
        data.append(length_3)

        for _ in range(length_3["Value"]):
            val_1 = next_int_len(4)
            data.append(val_1)

            val_2 = next_int_len(1)
            data.append(val_2)

    length = next_int_len(4)
    data.append(length)

    for _ in range(length["Value"]):
        val_1 = next_int_len(4)
        data.append(val_1)

        val_2 = next_int_len(1)
        data.append(val_2)

    length = next_int_len(4)
    data.append(length)

    for _ in range(length["Value"]):
        val_1 = next_int_len(4)
        data.append(val_1)

        val_2 = next_int_len(4)
        data.append(val_2)

    length = next_int_len(4)
    data.append(length)

    val_1 = next_int_len(4)
    data.append(val_1)

    for _ in range(length["Value"]):
        val_2 = next_int_len(8)
        data.append(val_2)

        val_3 = next_int_len(4)
        data.append(val_3)

    gv_56 = next_int_len(4)  # 0x38
    data.append(gv_56)

    val_1 = next_int_len(1)
    data.append(val_1)

    length = next_int_len(4)
    data.append(length)

    val_2 = next_int_len(4)
    data.append(val_2)

    for _ in range(length["Value"]):
        val_3 = next_int_len(1)
        data.append(val_3)

        val_4 = next_int_len(4)
        data.append(val_4)

    return data


def get_outbreaks() -> dict[int, Any]:
    chapters_count = next_int(4)
    outbreaks: dict[int, Any] = {}
    for _ in range(chapters_count):
        chapter_id = next_int(4)
        stages_count = next_int(4)
        chapter = {}
        for _ in range(stages_count):
            stage_id = next_int(4)
            outbreak_cleared_flag = next_int(1)
            chapter[stage_id] = outbreak_cleared_flag
        outbreaks[chapter_id] = chapter
    return outbreaks


def get_mission_data_maybe() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []

    length = next_int_len(4)
    data.append(length)

    for _ in range(length["Value"]):
        val_1 = next_int_len(4)
        data.append(val_1)

        val_2 = next_int_len(1)
        data.append(val_2)
    return data


def get_unlock_popups() -> tuple[list[tuple[int, int]], dict[str, int]]:
    """Get unlock popups and unlock flags"""

    length = next_int_len(4)
    val_1 = next_int_len(4)

    data: list[tuple[int, int]] = []

    for _ in range(length["Value"]):
        flag = next_int(1)

        popup_id = next_int(4)
        data.append((popup_id, flag))
    return data, val_1


def get_unknown_data():
    data: list[dict[str, int]] = []
    length = next_int_len(4)
    data.append(length)

    for _ in range(length["Value"]):
        length_2 = next_int_len(4)
        data.append(length_2)

        val_1 = next_int_len(4)
        data.append(val_1)

    unknown_val_2 = next_int_len(1)
    data.append(unknown_val_2)

    length = next_int_len(4)
    data.append(length)

    for _ in range(length["Value"]):
        val_1 = next_int_len(4)
        data.append(val_1)

        val_2 = next_int_len(1)
        data.append(val_2)

    return data


def get_cat_cannon_data() -> dict[int, dict[str, int]]:
    length = next_int(4)
    cannon_data: dict[int, dict[str, int]] = {}
    for _ in range(length):
        cannon = {"level": 0, "unlock_flag": 0}
        cannon_id = next_int(4)
        len_val = next_int(4)
        unlock_flag = next_int(4)
        level = next_int(4)
        extra_1 = 0
        extra_2 = 0
        if len_val == 4:
            extra_1 = next_int(4)
            extra_2 = next_int(4)
        cannon["level"] = level
        cannon["unlock_flag"] = unlock_flag
        cannon["len_val"] = len_val
        cannon["extra_1"] = extra_1
        cannon["extra_2"] = extra_2
        cannon_data[cannon_id] = cannon
    return cannon_data


def get_data_near_ht() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []

    val = next_int_len(1)
    data.append(val)

    for _ in range(val["Value"]):
        unknown_val_2 = next_int_len(3)
        data.append(unknown_val_2)

    unknown_val_3 = next_int_len(8)
    data.append(unknown_val_3)
    gv_64 = next_int_len(4)  # 0x40
    data.append(gv_64)

    length = next_int_len(4)
    data.append(length)

    val_1 = next_int_len(4)
    data.append(val_1)

    val_3 = {"Value": 0}
    for _ in range(length["Value"]):
        val_2 = next_int_len(4)
        data.append(val_2)

        val_3 = next_int_len(4)
        data.append(val_3)

    val_1 = next_int_len(4)
    data.append(val_1)

    val_4 = {"Value": 0}
    for _ in range(val_3["Value"]):
        length = next_int_len(4)
        data.append(length)

        for _ in range(length["Value"]):
            val_2 = next_int_len(4)
            data.append(val_2)

        val_4 = next_int_len(4)
        data.append(val_4)

    val_1 = next_int_len(4)
    data.append(val_1)

    for _ in range(val_4["Value"]):
        val_2 = next_int_len(1)
        data.append(val_2)

        val_3 = next_int_len(4)
        data.append(val_3)
    return data


def get_ht_it_data() -> dict[str, Any]:
    total = next_int(4)
    stars = next_int(4)

    current_data: dict[str, Any] = {}
    current_data = {"total": total, "stars": stars, "selected": []}

    for _ in range(total):
        for _ in range(stars):
            current_data["selected"].append(next_int(4))

    current_data["selected"] = list(helper.chunks(current_data["selected"], 4))

    total = next_int(4)
    stars = next_int(4)

    progress_data: dict[str, Any] = {}
    progress_data = {
        "total": total,
        "stars": stars,
        "clear_progress": [],
        "clear_amount": [],
        "unlock_next": [],
    }

    for _ in range(total):
        for _ in range(stars):
            progress_data["clear_progress"].append(next_int(4))

    progress_data["clear_progress"] = list(
        helper.chunks(progress_data["clear_progress"], 4)
    )

    total = next_int(4)
    stages = next_int(4)
    progress_data["stages"] = stages

    stars = next_int(4)

    clear_amount = get_length_data(4, 4, total * stages * stars)

    clear_amount = list(helper.chunks(clear_amount, stages * stars))

    clear_amount_sep: list[list[list[int]]] = []

    for clear_amount_val in clear_amount:
        sub_chapter_clears: list[list[int]] = []
        for j in range(stars):
            sub_chapter_clears.append(clear_amount_val[j::stars])
        clear_amount_sep.append(sub_chapter_clears)

    progress_data["clear_amount"] = clear_amount_sep

    data: list[dict[str, int]] = []
    length = next_int_len(4)
    data.append(length)

    length_2 = next_int_len(4)
    data.append(length_2)

    for _ in range(length["Value"]):
        for _ in range(length_2["Value"]):
            data.append(next_int_len(4))
    return {"data": data, "current": current_data, "progress": progress_data}


def get_mission_segment() -> dict[int, int]:
    missions: dict[int, int] = {}
    length = next_int(4)
    for _ in range(length):
        mission_id = next_int(4)
        mission_value = next_int(4)
        missions[mission_id] = mission_value
    return missions


def get_mission_data() -> dict[str, Any]:
    missions: dict[str, dict[int, int]] = {}
    missions["states"] = get_mission_segment()
    missions["requirements"] = get_mission_segment()
    missions["clear_types"] = get_mission_segment()
    missions["gamatoto"] = get_mission_segment()
    missions["nyancombo"] = get_mission_segment()
    missions["user_rank"] = get_mission_segment()
    missions["expiry"] = get_mission_segment()
    missions["preparing"] = get_mission_segment()
    return missions


def get_data_after_challenge() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []

    val_22 = next_int_len(4)
    data.append(val_22)

    gv_69 = next_int_len(4)  # 0x45
    data.append(gv_69)

    val_54 = next_int_len(4)
    data.append(val_54)

    val_118 = next_int_len(4)
    data.append(val_118)

    for _ in range(val_54["Value"]):
        val_15 = next_int_len(1)
        data.append(val_15)

        val_118 = next_int_len(4)
        data.append(val_118)

    val_54 = next_int_len(4)
    data.append(val_54)

    for _ in range(val_118["Value"]):
        val_65 = next_int_len(8)
        data.append(val_65)

        val_54 = next_int_len(4)
        data.append(val_54)

    return data


def get_data_after_tower() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []

    gv_66 = next_int_len(4)  # 0x42
    data.append(gv_66)

    data.append(next_int_len(4 * 2))
    data.append(next_int_len(1 * 3))
    data.append(next_int_len(4 * 3))
    data.append(next_int_len(1 * 3))
    data.append(next_int_len(1))
    data.append(next_int_len(8))

    val_54 = next_int_len(4)
    data.append(val_54)

    val_61 = next_int_len(4)
    data.append(val_61)

    for _ in range(val_54["Value"]):
        for _ in range(val_61["Value"]):
            val_22 = next_int_len(4)
            data.append(val_22)

    val_54 = next_int_len(4)
    data.append(val_54)

    val_61 = next_int_len(4)
    data.append(val_61)

    for _ in range(val_54["Value"]):
        for _ in range(val_61["Value"]):
            val_22 = next_int_len(4)
            data.append(val_22)

    val_54 = next_int_len(4)
    data.append(val_54)

    val_61 = next_int_len(4)
    data.append(val_61)

    val_57 = next_int_len(4)
    data.append(val_57)
    for _ in range(val_54["Value"]):
        for _ in range(val_61["Value"]):
            for _ in range(val_57["Value"]):
                val_22 = next_int_len(4)
                data.append(val_22)

    val_54 = next_int_len(4)
    data.append(val_54)

    val_61 = next_int_len(4)
    data.append(val_61)

    for _ in range(val_54["Value"]):
        for _ in range(val_61["Value"]):
            val_22 = next_int_len(4)
            data.append(val_22)

    val_54 = next_int_len(4)
    data.append(val_54)

    for _ in range(val_54["Value"] - 1):
        val_22 = next_int_len(4)
        data.append(val_22)

    return data


def get_uncanny_current() -> dict[str, Any]:
    total_subchapters = next_int(4)
    stages_per_subchapter = next_int(4)
    stars = next_int(4)
    if total_subchapters < 1:
        next_int(4)
        raise Exception("Invalid total subchapters")
    else:
        clear_progress = get_length_data(4, 4, total_subchapters * stars)
    clear_progress = list(helper.chunks(clear_progress, stars))

    return {
        "Clear": clear_progress,
        "total": total_subchapters,
        "stages": stages_per_subchapter,
        "stars": stars,
    }


def get_event_timed_scores() -> dict[str, Any]:
    total_subchapters = next_int(4)
    stages_per_subchapter = next_int(4)
    stars = next_int(4)

    score = get_length_data(4, 4, total_subchapters * stars * stages_per_subchapter)
    score = list(helper.chunks(score, stars * stages_per_subchapter))

    return {
        "Score": score,
        "total": total_subchapters,
        "stages": stages_per_subchapter,
        "stars": stars,
    }


def get_uncanny_progress(lengths: dict[str, Any]) -> dict[str, Any]:
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    clear_progress = get_length_data(4, 4, total * stars)
    clear_progress = list(helper.chunks(clear_progress, stars))

    clear_amount = get_length_data(4, 4, total * stages * stars)
    unlock_next = get_length_data(4, 4, total * stars)

    clear_amount = list(helper.chunks(clear_amount, stages * stars))
    unlock_next = list(helper.chunks(unlock_next, stars))

    clear_amount_sep: list[list[list[int]]] = []

    for clear_amount_val in clear_amount:
        sub_chapter_clears: list[list[int]] = []
        for j in range(stars):
            sub_chapter_clears.append(clear_amount_val[j::stars])
        clear_amount_sep.append(sub_chapter_clears)

    clear_amount = clear_amount_sep
    return {
        "Value": {
            "clear_progress": clear_progress,
            "clear_amount": clear_amount,
            "unlock_next": unlock_next,
        },
        "Lengths": lengths,
    }


def get_data_after_uncanny() -> dict[str, Any]:
    lengths = get_uncanny_current()
    return {"current": lengths, "progress": get_uncanny_progress(lengths)}


def get_gold_pass_data() -> dict[str, Any]:
    """Get gold pass related data"""

    data: dict[str, Any] = {}
    data["officer_id"] = next_int_len(4)
    data["renewal_times"] = next_int_len(4)
    data["start_date"] = get_double()
    data["expiry_date"] = get_double()
    data["unknown_2"] = get_length_doubles(length=2)
    data["start_date_2"] = get_double()
    data["expiry_date_2"] = get_double()
    data["unknown_3"] = get_double()
    data["flag_2"] = next_int_len(4)
    data["expiry_date_3"] = get_double()

    number_of_rewards = next_int(4)
    claimed_rewards: dict[int, int] = {}
    for _ in range(number_of_rewards):
        item_id = next_int(4)
        amount = next_int(4)
        claimed_rewards[item_id] = amount

    data["claimed_rewards"] = claimed_rewards
    data["unknown_4"] = next_int_len(8)
    data["unknown_5"] = next_int_len(1)
    data["unknown_6"] = next_int_len(1)

    return data


def get_talent_data() -> dict[int, list[dict[str, int]]]:
    total_cats = next_int(4)

    talents: dict[int, list[dict[str, int]]] = {}
    for _ in range(total_cats):
        cat_id = next_int(4)
        cat_data: list[dict[str, int]] = []

        number_of_talents = next_int(4)

        for _ in range(number_of_talents):
            talent_id = next_int(4)
            talent_level = next_int(4)
            talent = {"id": talent_id, "level": talent_level}
            cat_data.append(talent)
        talents[cat_id] = cat_data
    return talents


def get_medals() -> dict[str, Any]:
    medal_data_1 = get_length_data(2, 2)

    total_medals = next_int(2)
    medals = {}

    for _ in range(total_medals):
        medal_id = next_int(2)
        medal_flag = next_int(1)
        medals[medal_id] = medal_flag
    return {"medal_data_1": medal_data_1, "medal_data_2": medals}


def get_data_after_medals() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []
    data.append(next_int_len(1))

    val_2 = next_int_len(2)
    data.append(val_2)

    val_3 = next_int_len(2)
    data.append(val_3)

    for _ in range(val_2["Value"]):
        val_1 = next_int_len(1)
        data.append(val_1)

        val_3 = next_int_len(2)
        data.append(val_3)

    val_2 = next_int_len(2)
    data.append(val_2)

    val_6c = val_3
    for _ in range(val_6c["Value"]):
        val_2 = next_int_len(2)
        data.append(val_2)
        for _ in range(val_2["Value"]):
            val_3 = next_int_len(2)
            data.append(val_3)

            val_4 = next_int_len(2)
            data.append(val_4)

        val_2 = next_int_len(2)
        data.append(val_2)

    val_7c = val_2
    for _ in range(val_7c["Value"]):
        val_2 = next_int_len(2)
        data.append(val_2)

        val_12 = next_int_len(4)
        data.append(val_12)

    data.append(next_int_len(4))  # 90000

    data.append(next_int_len(4))
    data.append(next_int_len(4))
    data.append(next_int_len(8))
    data.append(next_int_len(4))  # 90100

    val_18 = next_int_len(2)
    data.append(val_18)
    for _ in range(val_18["Value"]):
        data.append(next_int_len(4))
        data.append(next_int_len(4))
        data.append(next_int_len(2))
        data.append(next_int_len(4))
        data.append(next_int_len(4))
        data.append(next_int_len(4))
        data.append(next_int_len(2))

    val_18 = next_int_len(2)
    data.append(val_18)
    for _ in range(val_18["Value"]):
        val_32 = next_int_len(4)
        data.append(val_32)

        val_48 = next_int_len(8)
        data.append(val_48)

    return data


def get_data_after_after_leadership(dst: bool) -> list[dict[str, int]]:
    data: list[dict[str, int]] = []
    data.append(next_int_len(4))
    if not dst:
        data.append(next_int_len(5))

    if dst:
        data.append(next_int_len(12))
    else:
        data.append(next_int_len(7))
    return data


def get_legend_quest_current() -> dict[str, Any]:
    total_subchapters = next_int(1)
    stages_per_subchapter = next_int(1)
    stars = next_int(1)

    clear_progress = get_length_data(4, 1, total_subchapters * stars)
    clear_progress = list(helper.chunks(clear_progress, stars))

    return {
        "Clear": clear_progress,
        "total": total_subchapters,
        "stages": stages_per_subchapter,
        "stars": stars,
    }


def get_legend_quest_progress(lengths: dict[str, Any]):
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    clear_progress = get_length_data(4, 1, total * stars)
    clear_progress = list(helper.chunks(clear_progress, stars))
    clear_amount = get_length_data(4, 2, total * stars * stages)
    tries = get_length_data(4, 2, total * stars * stages)

    unlock_next = get_length_data(4, 1, total * stars)
    unlock_next = list(helper.chunks(unlock_next, stars))

    clear_amount = list(helper.chunks(clear_amount, stages * stars))
    tries = list(helper.chunks(tries, stages * stars))

    clear_amount_sep: list[list[list[int]]] = []
    stage_ids_sep: list[list[list[int]]] = []

    for clear_amount_val in clear_amount:
        sub_chapter_clears: list[list[int]] = []
        for j in range(stars):
            sub_chapter_clears.append(clear_amount_val[j::stars])
        clear_amount_sep.append(sub_chapter_clears)
    clear_amount = clear_amount_sep

    for stage_id_val in tries:
        sub_chapter_ids: list[list[int]] = []
        for j in range(stars):
            sub_chapter_ids.append(stage_id_val[j::stars])
        stage_ids_sep.append(sub_chapter_ids)
    tries = stage_ids_sep

    return {
        "Value": {
            "clear_progress": clear_progress,
            "clear_amount": clear_amount,
            "tries": tries,
            "unlock_next": unlock_next,
        },
        "Lengths": lengths,
    }


def get_data_after_leadership() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []

    data.append(next_int_len(2))
    data.append(next_int_len(1))
    data.append(next_int_len(4))  # 80600

    val_54 = next_int_len(4)
    data.append(val_54)

    if val_54["Value"] > 0:
        val_118 = next_int_len(4)
        data.append(val_118)

        val_55 = next_int_len(4)
        data.append(val_55)

        for _ in range(val_55["Value"]):
            val_61 = next_int_len(4)
            data.append(val_61)

        for _ in range(val_54["Value"] - 1):
            val_61 = next_int_len(4)
            data.append(val_61)

            val_61 = next_int_len(4)
            data.append(val_61)
    return data


def get_gauntlet_current() -> dict[str, Any]:
    total_subchapters = next_int(2)
    stages_per_subchapter = next_int(1)
    stars = next_int(1)

    clear_progress = get_length_data(4, 1, total_subchapters * stars)
    clear_progress = list(helper.chunks(clear_progress, stars))

    return {
        "Clear": clear_progress,
        "total": total_subchapters,
        "stages": stages_per_subchapter,
        "stars": stars,
    }


def get_gauntlet_progress(
    lengths: dict[str, Any], unlock: bool = True
) -> dict[str, Any]:
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    clear_progress = get_length_data(4, 1, total * stars)
    clear_progress = list(helper.chunks(clear_progress, stars))

    clear_amount = get_length_data(4, 2, total * stages * stars)
    unlock_next = []
    if unlock:
        unlock_next = get_length_data(4, 1, total * stars)
        unlock_next = list(helper.chunks(unlock_next, stars))

    clear_amount = list(helper.chunks(clear_amount, stages * stars))

    clear_amount_sep: list[list[list[int]]] = []

    for clear_amount_val in clear_amount:
        sub_chapter_clears: list[list[int]] = []
        for j in range(stars):
            sub_chapter_clears.append(clear_amount_val[j::stars])
        clear_amount_sep.append(sub_chapter_clears)

    clear_amount = clear_amount_sep
    return {
        "Value": {
            "clear_progress": clear_progress,
            "clear_amount": clear_amount,
            "unlock_next": unlock_next,
        },
        "Lengths": lengths,
    }


class ClearedSlots:
    class Slot:
        class Cat:
            def __init__(self, cat_id: int, cat_form: int):
                self.cat_id = cat_id
                self.cat_form = cat_form

            def to_dict(self) -> dict[str, Any]:
                return {
                    "cat_id": self.cat_id,
                    "cat_form": self.cat_form,
                }

            @staticmethod
            def from_dict(data: dict[str, Any]) -> "ClearedSlots.Slot.Cat":
                return ClearedSlots.Slot.Cat(data["cat_id"], data["cat_form"])

        def __init__(self, cats: list[Cat], slot_index: int, separator: int):
            self.cats = cats
            self.slot_index = slot_index
            self.separator = separator

        def to_dict(self) -> dict[str, Any]:
            return {
                "cats": [cat.to_dict() for cat in self.cats],
                "slot_index": self.slot_index,
                "separator": self.separator,
            }

        @staticmethod
        def from_dict(data: dict[str, Any]):
            return ClearedSlots.Slot(
                [ClearedSlots.Slot.Cat.from_dict(cat) for cat in data["cats"]],
                data["slot_index"],
                data["separator"],
            )

    class StageSlot:
        class Stage:
            def __init__(self, stage_id: int):
                self.stage_id = stage_id

            def to_dict(self) -> dict[str, Any]:
                return {
                    "stage_id": self.stage_id,
                }

            @staticmethod
            def from_dict(data: dict[str, Any]) -> "ClearedSlots.StageSlot.Stage":
                return ClearedSlots.StageSlot.Stage(data["stage_id"])

        def __init__(self, slot_index: int, stages: list[Stage]):
            self.slot_index = slot_index
            self.stages = stages

        def to_dict(self) -> dict[str, Any]:
            return {
                "slot_index": self.slot_index,
                "stages": [stage.to_dict() for stage in self.stages],
            }

        @staticmethod
        def from_dict(data: dict[str, Any]) -> "ClearedSlots.StageSlot":
            return ClearedSlots.StageSlot(
                data["slot_index"],
                [
                    ClearedSlots.StageSlot.Stage.from_dict(stage)
                    for stage in data["stages"]
                ],
            )

    def __init__(self, slots: list[Slot], slot_stages: list[StageSlot], end_index: int):
        self.slots = slots
        self.slot_stages = slot_stages
        self.end_index = end_index

    def to_dict(self) -> dict[str, Any]:
        return {
            "slots": [slot.to_dict() for slot in self.slots],
            "slot_stages": [stage.to_dict() for stage in self.slot_stages],
            "end_index": self.end_index,
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> "ClearedSlots":
        return ClearedSlots(
            [ClearedSlots.Slot.from_dict(slot) for slot in data["slots"]],
            [ClearedSlots.StageSlot.from_dict(stage) for stage in data["slot_stages"]],
            data["end_index"],
        )


def get_enigma_stages() -> dict[str, Any]:
    """
    Gets the enigma stages

    Returns:
        dict[str, Any]: The enigma stages
    """
    enigma_data: dict[str, Any] = {}
    enigma_data["energy_since_1"] = next_int(4)
    enigma_data["energy_since_2"] = next_int(4)
    enigma_data["enigma_level"] = next_int(1)
    enigma_data["unknown_2"] = next_int(1)
    enigma_data["unknown_3"] = next_int(1)

    total_stages = next_int(1)
    stages: list[dict[str, Any]] = []
    for _ in range(total_stages):
        data = {}
        data["level"] = next_int(4)  # 0 = inferior, 1 = normal, 2 = superior
        data["stage_id"] = next_int(4)
        data["decoding_status"] = next_int(
            1
        )  # 0 = not decoded, 1 = decoded, 2 = revealed
        data["start_time"] = get_double()
        stages.append(data)
    enigma_data["stages"] = stages
    return enigma_data


def get_cleared_slots() -> tuple[dict[str, Any], list[dict[str, int]]]:
    """
    Returns the line ups of the cleared stages

    Returns:
        dict[str, Any]: The line ups of the cleared stages
    """
    total_slots = next_int(2)
    index = next_int(2)
    slots: list[ClearedSlots.Slot] = []

    for _ in range(total_slots):
        cats: list[ClearedSlots.Slot.Cat] = []
        for _ in range(10):
            cat_id = next_int(2)
            cat_form = next_int(1)
            cat_data = ClearedSlots.Slot.Cat(cat_id, cat_form)
            cats.append(cat_data)
        separator = next_int(3)
        slot = ClearedSlots.Slot(cats, index, separator)
        index = next_int(2)
        slots.append(slot)

    cleared_slot_data: list[ClearedSlots.StageSlot] = []
    index_2 = next_int(2)
    for _ in range(index):
        total_stages = next_int(2)
        stages: list[ClearedSlots.StageSlot.Stage] = []
        for _ in range(total_stages):
            stage_id = next_int(4)
            stage = ClearedSlots.StageSlot.Stage(stage_id)
            stages.append(stage)
        stages_data = ClearedSlots.StageSlot(index_2, stages)
        index_2 = next_int(2)
        cleared_slot_data.append(stages_data)

    data_2: list[dict[str, int]] = []
    data_2.append({"Value": index_2, "Length": 2})
    for _ in range(index_2):
        val_18 = next_int_len(2)
        data_2.append(val_18)

        val_4 = next_int_len(1)
        data_2.append(val_4)
    data_2.append(next_int_len(4))  # 90400

    cleared_slots = ClearedSlots(slots, cleared_slot_data, index)

    return cleared_slots.to_dict(), data_2


def get_data_after_gauntlets() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []

    data.append(next_int_len(4 * 2))
    data.append(next_int_len(1 * 3))

    val_4 = next_int_len(1)
    data.append(val_4)
    for _ in range(val_4["Value"]):
        data.append(next_int_len(4))
        data.append(next_int_len(4))
        data.append(next_int_len(1))
        data.append(next_int_len(8))

    return data


def get_data_after_orbs() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []
    val_31 = next_int_len(2)
    data.append(val_31)
    for _ in range(val_31["Value"]):
        val_18 = next_int_len(2)
        data.append(val_18)

        val_5 = next_int_len(1)
        data.append(val_5)

        for _ in range(val_5["Value"]):
            val_6 = next_int_len(1)
            data.append(val_6)

            val_18 = next_int_len(2)
            data.append(val_18)
    data.append(next_int_len(1))
    data.append(next_int_len(4))  # 90700

    length = next_int_len(2)
    data.append(length)
    for _ in range(length["Value"]):
        data.append(next_int_len(4))

    data.append(next_int_len(1 * 10))
    data.append(next_int_len(4))  # 90800

    data.append(next_int_len(1))
    return data


def get_cat_shrine_data() -> dict[str, Any]:
    """
    Gets the cat shrine data

    Returns:
        dict[str, Any]: The cat shrine data
    """
    stamp_1 = get_double()
    stamp_2 = get_double()
    shrine_gone = next_int(1)
    flags: list[int] = get_length_data(1, 1)
    xp_offering = next_int(4)
    return {
        "flags": flags,
        "xp_offering": xp_offering,
        "shrine_gone": shrine_gone,
        "stamp_1": stamp_1,
        "stamp_2": stamp_2,
    }


def get_slot_names(save_stats: dict[str, Any]) -> list[str]:
    total_slots = len(save_stats["slots"])
    if save_stats["game_version"]["Value"] >= 110600:
        total_slots = next_int(1)
    names: list[str] = []
    for _ in range(total_slots):
        names.append(get_utf8_string())
    return names


def get_talent_orbs(game_version: dict[str, Any]) -> dict[int, int]:
    talent_orb_data: dict[int, int] = {}

    total_orbs = next_int(2)
    for _ in range(total_orbs):
        orb_id = next_int(2)
        if game_version["Value"] < 110400:
            amount = next_int(1)
        else:
            amount = next_int(2)
        talent_orb_data[orb_id] = amount

    return talent_orb_data


def data_after_after_gauntlets() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []
    data.append(next_int_len(1))
    data.append(next_int_len(8 * 2))
    data.append(next_int_len(4))
    data.append(next_int_len(1 * 2))
    data.append(next_int_len(8 * 2))
    data.append(next_int_len(4))  # 90500
    return data


def get_data_near_end_after_shards() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []
    data.append(next_int_len(1))
    data.append(next_int_len(4))  # 100600

    val_2 = next_int_len(2)
    data.append(val_2)

    val_3 = next_int_len(2)
    data.append(val_3)

    for _ in range(val_2["Value"]):
        val_1 = next_int_len(1)
        data.append(val_1)

        val_3 = next_int_len(2)
        data.append(val_3)
    val_6c = val_3

    val_2 = next_int_len(2)
    data.append(val_2)

    for _ in range(val_6c["Value"]):
        val_2 = next_int_len(2)
        data.append(val_2)

        for _ in range(val_2["Value"]):
            val_3 = next_int_len(2)
            data.append(val_3)

            val_4 = next_int_len(2)
            data.append(val_4)

        val_2 = next_int_len(2)
        data.append(val_2)
    val_7c = val_2
    for _ in range(val_7c["Value"]):
        val_2 = next_int_len(2)
        data.append(val_2)

        val_12 = next_int_len(4)
        data.append(val_12)
    return data


def get_data_near_end() -> list[dict[str, int]]:
    data: list[dict[str, int]] = []
    val_5 = next_int_len(1)
    data.append(val_5)
    if 0 < val_5["Value"]:
        val_33 = next_int_len(4)
        data.append(val_33)
        if val_5["Value"] != 1:
            val_33 = next_int_len(4)
            data.append(val_33)
            if val_5["Value"] != 2:
                val_32 = val_5["Value"] + 2
                for _ in range(val_32):
                    data.append(next_int_len(4))

    data.append(next_int_len(1))
    data.append(next_int_len(4))  # 100400
    data.append(next_int_len(8))
    return data


def get_aku() -> dict[str, Any]:
    total = next_int(2)
    stages = next_int(1)
    stars = next_int(1)
    return get_gauntlet_progress(
        {"total": total, "stages": stages, "stars": stars}, False
    )


def get_data_after_aku() -> list[dict[str, int]]:
    data_1: list[dict[str, int]] = []

    val_6 = next_int_len(2)
    data_1.append(val_6)

    val_7 = next_int_len(2)
    data_1.append(val_7)

    for _ in range(val_6["Value"]):
        val_7 = next_int_len(2)
        data_1.append(val_7)

        for _ in range(val_7["Value"]):
            data_1.append(next_int_len(2))
        val_7 = next_int_len(2)
        data_1.append(val_7)

    val_4c = val_7
    for _ in range(val_4c["Value"]):
        data_1.append(next_int_len(2))
        data_1.append(next_int_len(8))

    val_5 = next_int_len(2)
    data_1.append(val_5)

    for _ in range(val_5["Value"]):
        data_1.append(next_int_len(2))
        data_1.append(next_int_len(8))

    data_1.append(next_int_len(1))
    return data_1


def get_data_near_end_after_aku() -> list[dict[str, int]]:
    data_2: list[dict[str, int]] = []
    val_4 = next_int_len(2)
    data_2.append(val_4)

    for _ in range(val_4["Value"]):
        data_2.append(next_int_len(4))
        data_2.append(next_int_len(1))
        data_2.append(next_int_len(1))
    return data_2


def exit_parser(save_stats: dict[str, Any]) -> dict[str, Any]:
    save_stats["hash"] = get_utf8_string(32)
    return save_stats


def check_gv(save_stats: dict[str, Any], game_version: int) -> dict[str, Any]:
    if save_stats["game_version"]["Value"] < game_version:
        save_stats = exit_parser(save_stats)
        save_stats["exit"] = True
        save_stats["extra_data"] = next_int_len(0)
    else:
        save_stats["exit"] = False
    return save_stats


def get_play_time() -> dict[str, Any]:
    raw_val = next_int_len(4)
    frames = raw_val["Value"]

    play_time_data = helper.frames_to_time(frames)

    return play_time_data


def start_parse(save_data: bytes, country_code: str) -> dict[str, Any]:
    """Start the parser and handle any exceptions."""

    try:
        save_stats = parse_save(save_data, country_code)
    except Exception:  # pylint: disable=broad-except
        helper.colored_text(
            f"\nError: An error has occurred while parsing your save data (address = {address}):",
            base=helper.RED,
        )
        traceback.print_exc()
        game_version = get_game_version(save_data)
        if game_version < 110000:
            helper.colored_text(
                f"\nThis save is from before &11.0.0& (current save version is &{helper.gv_to_str(game_version)}&), so this is likely the cause for the issue. &The save editor is not designed to work with saves from before 11.0.0&"
            )
        else:
            helper.colored_text(
                "\nPlease report this to &#bug-reports&, and/or &dm me your save& on discord"
            )
        helper.exit_editor()
    return save_stats


def get_game_version(save_data: bytes) -> int:
    """Get the game version from the save data."""

    return convert_little(save_data[0:3])


def find_date() -> int:
    """Find the date of the save, used because for some reason in some saves there is like 40 zero bytes before the main save data"""
    for _ in range(100):
        val = next_int(4)
        if val >= 2000 and val <= 3000:
            return address - 4
    raise Exception("Could not find date")


def get_dst(save_data: bytes, offset: int) -> bool:
    """Get if the save has daylight savings from the save data, this is used to handle jp differences."""

    dst = False
    if save_data[offset] >= 15 and save_data[offset] <= 20:
        dst = True
    elif save_data[offset - 1] >= 15 and save_data[offset - 1] <= 20:
        dst = False  # Offset in jp due to no dst
    return dst


def get_double() -> float:
    """Get a double from the save data."""

    if save_data_g is None:
        raise Exception("No save data loaded")
    data = save_data_g[address : address + 8]
    val = struct.unpack("d", data)[0]
    set_address(address + 8)
    return val


def get_110800_data() -> list[dict[str, int]]:
    """
    Get the data from 11.7.0

    Returns:
        list[dict[str, int]]: The data
    """
    data: list[dict[str, int]] = []

    u_var_38 = next_int_len(1)
    data.append(u_var_38)

    return data


def get_110800_data_2() -> list[dict[str, int]]:
    """
    Get the data from 11.7.0

    Returns:
        list[dict[str, int]]: The data
    """
    data: list[dict[str, int]] = []

    u_var_38 = next_int_len(1)
    data.append(u_var_38)

    u_var_38 = next_int_len(1)
    data.append(u_var_38)

    return data


def get_110700_data() -> list[dict[str, int]]:
    """
    Get the data from 110600

    Returns:
        list[dict[str, int]]: The data
    """
    data: list[dict[str, int]] = []

    i_var_32 = next_int_len(4)
    data.append(i_var_32)

    for _ in range(i_var_32["Value"]):
        pi_var_33 = next_int_len(4)
        data.append(pi_var_33)

        f_var_54 = next_int_len(8)
        data.append(f_var_54)

        f_var_54 = next_int_len(8)
        data.append(f_var_54)

    return data


def get_login_bonuses() -> dict[int, int]:
    """
    Get the login bonuses

    Returns:
        dict[int, int]: The login bonuses
    """
    length = next_int(4)
    data: dict[int, int] = {}
    for _ in range(length):
        login_id = next_int(4)
        data[login_id] = next_int(4)
    return data


def get_tower_item_obtained() -> list[list[int]]:
    total_stars = next_int(4)
    total_stages = next_int(4)
    data: list[list[int]] = []
    for _ in range(total_stars):
        star_data: list[int] = []
        for _ in range(total_stages):
            star_data.append(next_int(1))
        data.append(star_data)
    return data


def get_dict(
    key_type: type, value_type: type, length: Optional[int] = None
) -> dict[Any, Any]:
    if length is None:
        length = next_int(4)
    data: dict[Any, Any] = {}
    for _ in range(length):
        if key_type == int:
            key = next_int(4)
        else:
            raise Exception("Invalid key type")
        if value_type == int:
            data[key] = next_int(4)
        elif value_type == str:
            data[key] = get_utf8_string()
        elif value_type == bool:
            data[key] = next_int(1) == 1
        else:
            raise Exception("Invalid value type")
    return data


def parse_save(
    save_data: bytes,
    country_code: Union[str, None],
    dst: Optional[bool] = None,
) -> dict[str, Any]:
    """Parse the save data."""

    set_address(0)
    global save_data_g
    save_data_g = save_data
    save_stats: dict[str, Any] = {}
    save_stats["editor_version"] = updater.get_local_version()

    save_stats["game_version"] = next_int_len(4)
    save_stats["version"] = country_code

    save_stats["unknown_1"] = next_int_len(1)

    save_stats["mute_music"] = next_int_len(1)
    save_stats["mute_sound_effects"] = next_int_len(1)

    save_stats["cat_food"] = next_int_len(4)
    save_stats["current_energy"] = next_int_len(4)

    old_address = address
    new_address = find_date()
    set_address(old_address)
    extra = new_address - old_address
    save_stats["extra_time_data"] = next_int_len(extra)

    if dst is None:
        dst = get_dst(save_data, address + 118)
    save_stats["dst"] = dst
    if (
        save_stats["version"] == "jp"
        and dst
        or save_stats["version"] != "jp"
        and not dst
    ):
        helper.colored_text(
            "Warning: DST detected is not correct for this save version, this may cause issues with save parsing.",
            helper.RED,
        )

    data = get_time_data_skip(save_stats["dst"])

    save_stats["time"] = data["time"]
    save_stats["dst_val"] = data["dst"]
    save_stats["time_stamp"] = data["time_stamp"]
    save_stats["duplicate_time"] = data["duplicate"]

    save_stats["unknown_flags_1"] = get_length_data(length=3)
    save_stats["upgrade_state"] = next_int_len(4)
    save_stats["xp"] = next_int_len(4)

    save_stats["tutorial_cleared"] = next_int_len(4)
    save_stats["unknown_flags_2"] = get_length_data(length=12)
    save_stats["unknown_flag_1"] = next_int_len(1)
    save_stats["slots"] = get_equip_slots()

    save_stats["cat_stamp_current"] = next_int_len(4)

    save_stats["cat_stamp_collected"] = get_length_data(length=30)
    save_stats["unknown_2"] = next_int_len(4)
    save_stats["daily_reward_flag"] = next_int_len(4)
    save_stats["unknown_116"] = get_length_data(length=10)

    save_stats["story_chapters"] = get_main_story_levels()
    save_stats["treasures"] = get_treasures()
    try:
        save_stats["enemy_guide"] = get_length_data()
    except Exception:
        return parse_save(save_data, country_code, not dst)
    if len(save_stats["enemy_guide"]) == 0:
        return parse_save(save_data, country_code, not dst)
    save_stats["cats"] = get_length_data()
    save_stats["cat_upgrades"] = get_cat_upgrades()
    save_stats["current_forms"] = get_length_data()

    save_stats["blue_upgrades"] = get_blue_upgrades()

    save_stats["menu_unlocks"] = get_length_data()
    save_stats["new_dialogs_1"] = get_length_data()

    save_stats["battle_items"] = get_length_data(4, 4, 6)
    save_stats["new_dialogs_2"] = get_length_data()
    save_stats["unknown_6"] = next_int_len(4)
    save_stats["unknown_7"] = get_length_data(length=21)

    save_stats["lock_item"] = next_int_len(1)
    save_stats["locked_items"] = get_length_data(1, 1, 6)
    save_stats["second_time"] = get_time_data(save_stats["dst"])

    save_stats["unknown_8"] = get_length_data(length=50)
    save_stats["third_time"] = get_time_data(save_stats["dst"])

    save_stats["unknown_9"] = next_int_len(6 * 4)

    save_stats["thirty2_code"] = get_utf8_string()
    save_stats["unknown_10"] = load_bonus_hash()
    save_stats["unknown_11"] = get_length_data(length=4)
    save_stats["normal_tickets"] = next_int_len(4)
    save_stats["rare_tickets"] = next_int_len(4)
    save_stats["gatya_seen_cats"] = get_length_data()
    save_stats["unknown_12"] = get_length_data(length=10)
    length = next_int(2)
    cat_storage_len = True
    if length != 128:
        skip(-2)
        cat_storage_len = False
        length = 100

    cat_storage_id = get_length_data(2, 4, length)
    cat_storage_type = get_length_data(2, 4, length)
    save_stats["cat_storage"] = {
        "ids": cat_storage_id,
        "types": cat_storage_type,
        "len": cat_storage_len,
    }
    current_sel = get_event_stages_current()
    save_stats["event_current"] = current_sel

    save_stats["event_stages"] = get_event_stages(current_sel)

    save_stats["unknown_15"] = get_length_data(length=38)
    save_stats["unit_drops"] = get_length_data()
    save_stats["rare_gacha_seed"] = next_int_len(4)
    save_stats["unknown_17"] = next_int_len(12)
    save_stats["unknown_18"] = next_int_len(4)

    save_stats["fourth_time"] = get_time_data(save_stats["dst"])
    save_stats["unknown_105"] = get_length_data(length=5)
    save_stats["unknown_107"] = get_length_data(separator=1, length=3)

    if save_stats["dst"]:
        save_stats["unknown_110"] = get_utf8_string()
    else:
        save_stats["unknown_110"] = ""

    total_strs = next_int(4)
    unknown_108: list[str] = []
    for _ in range(total_strs):
        unknown_108.append(get_utf8_string())
    save_stats["unknown_108"] = unknown_108

    if save_stats["dst"]:
        save_stats["time_stamps"] = get_length_doubles(length=3)

        length = next_int(4)
        strs: list[str] = []
        for _ in range(length):
            strs.append(get_utf8_string())
        save_stats["unknown_112"] = strs
        save_stats["energy_notice"] = next_int_len(1)
        save_stats["game_version_2"] = next_int_len(4)
    else:
        save_stats["time_stamps"] = [0, 0, 0]
        save_stats["unknown_112"] = []
        save_stats["energy_notice"] = generate_empty_len(1)
        save_stats["game_version_2"] = generate_empty_len(4)

    save_stats["unknown_111"] = next_int_len(4)
    save_stats["unlocked_slots"] = next_int_len(1)

    length_1 = next_int(4)
    length_2 = next_int(4)
    unknown_20: dict[str, Any] = {}
    unknown_20 = {"Value": get_length_data(4, 4, length_1 * length_2)}
    unknown_20["Length_1"] = length_1
    unknown_20["Length_2"] = length_2
    save_stats["unknown_20"] = unknown_20

    save_stats["time_stamps_2"] = get_length_doubles(length=4)

    save_stats["trade_progress"] = next_int_len(4)

    if save_stats["dst"]:
        save_stats["time_stamps_2"].append(get_double())
        save_stats["unknown_24"] = generate_empty_len(4)
    else:
        save_stats["unknown_24"] = next_int_len(4)

    save_stats["catseye_related_data"] = get_cat_upgrades()
    save_stats["unknown_22"] = get_length_data(length=11)
    save_stats["user_rank_rewards"] = get_length_data(4, 1)

    if not save_stats["dst"]:
        save_stats["time_stamps_2"].append(get_double())

    save_stats["unlocked_forms"] = get_length_data()
    save_stats["transfer_code"] = get_utf8_string()
    save_stats["confirmation_code"] = get_utf8_string()
    save_stats["transfer_flag"] = next_int_len(1)

    lengths = [next_int(4), next_int(4), next_int(4)]
    length = lengths[0] * lengths[1] * lengths[2]

    save_stats["stage_data_related_1"] = {
        "Value": get_length_data(4, 1, length),
        "Lengths": lengths,
    }

    save_stats["event_timed_scores"] = get_event_timed_scores()
    save_stats["inquiry_code"] = get_utf8_string()
    save_stats["play_time"] = get_play_time()

    save_stats["unknown_25"] = next_int_len(1)

    save_stats["backup_state"] = next_int_len(4)

    if save_stats["dst"]:
        save_stats["unknown_119"] = next_int_len(1)
    else:
        save_stats["unknown_119"] = generate_empty_len(1)

    save_stats["gv_44"] = next_int_len(4)

    save_stats["unknown_120"] = next_int_len(4)

    save_stats["itf_timed_scores"] = list(
        helper.chunks(get_length_data(4, 4, 51 * 3), 51)
    )

    save_stats["unknown_27"] = next_int_len(4)
    save_stats["cat_related_data_1"] = get_length_data()
    save_stats["unknown_28"] = next_int_len(1)

    save_stats["gv_45"] = next_int_len(4)
    save_stats["gv_46"] = next_int_len(4)

    save_stats["unknown_29"] = next_int_len(4)
    save_stats["lucky_tickets_1"] = get_length_data()
    save_stats["unknown_32"] = get_length_data()

    save_stats["gv_47"] = next_int_len(4)
    save_stats["gv_48"] = next_int_len(4)

    if not save_stats["dst"]:
        save_stats["energy_notice"] = next_int_len(1)
    save_stats["account_created_time_stamp"] = get_double()

    save_stats["unknown_35"] = get_length_data()
    save_stats["unknown_36"] = next_int_len(15)

    save_stats["user_rank_popups"] = next_int_len(3)

    save_stats["unknown_37"] = next_int_len(1)

    save_stats["gv_49"] = next_int_len(4)
    save_stats["gv_50"] = next_int_len(4)
    save_stats["gv_51"] = next_int_len(4)
    save_stats["cat_guide_collected"] = get_length_data(4, 1)

    save_stats["gv_52"] = next_int_len(4)

    save_stats["time_stamps_3"] = get_length_doubles(length=5)

    save_stats["cat_fruit"] = get_length_data()
    save_stats["cat_related_data_3"] = get_length_data()
    save_stats["catseye_cat_data"] = get_length_data()
    save_stats["catseyes"] = get_length_data()
    save_stats["catamins"] = get_length_data()

    save_stats["gamatoto_time_left"] = helper.seconds_to_time(int(get_double()))
    save_stats["gamatoto_exclamation"] = next_int_len(1)
    save_stats["gamatoto_xp"] = next_int_len(4)
    save_stats["gamamtoto_destination"] = next_int_len(4)
    save_stats["gamatoto_recon_length"] = next_int_len(4)

    save_stats["unknown_43"] = next_int_len(4)

    save_stats["gamatoto_complete_notification"] = next_int_len(4)

    save_stats["unknown_44"] = get_length_data(4, 1)
    save_stats["unknown_45"] = get_length_data(4, 12 * 4)
    save_stats["gv_53"] = next_int_len(4)

    save_stats["helpers"] = get_length_data()

    save_stats["unknown_47"] = next_int_len(1)

    save_stats["gv_54"] = next_int_len(4)

    save_stats["purchases"] = get_purchase_receipts()
    save_stats["gv_54"] = next_int_len(4)
    save_stats["gamatoto_skin"] = next_int_len(4)
    save_stats["platinum_tickets"] = next_int_len(4)

    save_stats["login_bonuses"] = get_login_bonuses()
    save_stats["unknown_49"] = next_int_len(16)
    save_stats["announcment"] = get_length_data(length=32)

    save_stats["backup_counter"] = next_int_len(4)

    save_stats["unknown_131"] = get_length_data(length=3)
    save_stats["gv_55"] = next_int_len(4)

    save_stats["unknown_51"] = next_int_len(1)

    save_stats["unknown_113"] = get_data_before_outbreaks()

    save_stats["dojo_data"] = get_dojo_data_maybe()
    save_stats["dojo_item_lock"] = next_int_len(1)
    save_stats["dojo_locks"] = get_length_data(1, 1, 2)

    save_stats["unknown_114"] = next_int_len(4)
    save_stats["gv_58"] = next_int_len(4)  # 0x3a
    save_stats["unknown_115"] = next_int_len(8)

    save_stats["outbreaks"] = get_outbreaks()

    save_stats["unknown_52"] = get_double()
    save_stats["item_schemes"] = {}
    save_stats["item_schemes"]["to_obtain_ids"] = get_length_data()
    save_stats["item_schemes"]["received_ids"] = get_length_data()

    save_stats["current_outbreaks"] = get_outbreaks()

    save_stats["unknown_55"] = get_mission_data_maybe()

    save_stats["time_stamp_4"] = get_double()
    save_stats["gv_60"] = next_int_len(4)

    save_stats["unknown_117"] = get_unknown_data()

    save_stats["gv_61"] = next_int_len(4)
    data = get_unlock_popups()
    save_stats["unlock_popups"] = data[0]

    save_stats["unknown_118"] = data[1]

    save_stats["base_materials"] = get_length_data()

    save_stats["unknown_56"] = next_int_len(8)
    save_stats["unknown_57"] = next_int_len(1)
    save_stats["unknown_58"] = next_int_len(4)

    save_stats["engineers"] = next_int_len(4)
    save_stats["ototo_cannon"] = get_cat_cannon_data()

    save_stats["unknown_59"] = get_data_near_ht()

    save_stats["tower"] = get_ht_it_data()
    save_stats["missions"] = get_mission_data()
    save_stats["tower_item_obtained"] = get_tower_item_obtained()
    save_stats["unknown_61"] = get_data_after_tower()

    save_stats["challenge"] = {"Score": next_int_len(4), "Cleared": next_int_len(1)}

    save_stats["gv_67"] = next_int_len(4)  # 0x43

    save_stats["weekly_event_missions"] = get_dict(int, bool)
    save_stats["won_dojo_reward"] = next_int_len(1)
    save_stats["event_flag_update_flag"] = next_int_len(1)

    save_stats["gv_68"] = next_int_len(4)  # 0x44

    save_stats["completed_one_level_in_chapter"] = get_dict(int, int)
    save_stats["displayed_cleared_limit_text"] = get_dict(int, bool)
    save_stats["event_start_dates"] = get_dict(int, int)
    save_stats["stages_beaten_twice"] = get_length_data()

    save_stats["unknown_102"] = get_data_after_challenge()

    lengths = get_uncanny_current()
    save_stats["uncanny_current"] = lengths
    save_stats["uncanny"] = get_uncanny_progress(lengths)

    total = lengths["total"]
    save_stats["unknown_62"] = next_int_len(4)
    save_stats["unknown_63"] = get_length_data(length=total)

    save_stats["unknown_64"] = get_data_after_uncanny()

    total = save_stats["unknown_64"]["progress"]["Lengths"]["total"]
    save_stats["unknown_65"] = next_int_len(4)
    val_61 = save_stats["unknown_65"]

    save_stats["unknown_66"] = []
    unknown_66: list[Any] = []
    for _ in range(total):
        val_61 = next_int_len(4)
        unknown_66.append(val_61)
    save_stats["unknown_66"] = unknown_66

    val_54 = 0x37
    if val_61["Value"] < 0x38:
        val_54 = val_61["Value"]

    save_stats["lucky_tickets_2"] = get_length_data(length=val_54)

    save_stats["unknown_67"] = []
    if 0x37 < val_61["Value"]:
        save_stats["unknown_67"] = get_length_data(4, 4, val_61["Value"])

    save_stats["unknown_68"] = next_int_len(1)

    save_stats["gv_77"] = next_int_len(4)  # 0x4d

    save_stats["gold_pass"] = get_gold_pass_data()

    save_stats["talents"] = get_talent_data()
    save_stats["np"] = next_int_len(4)

    save_stats["unknown_70"] = next_int_len(1)

    save_stats["gv_80000"] = next_int_len(4)  # 80000

    save_stats["unknown_71"] = next_int_len(1)

    save_stats["leadership"] = next_int_len(2)
    save_stats["officer_pass_cat_id"] = next_int_len(2)
    save_stats["officer_pass_cat_form"] = next_int_len(2)

    save_stats["gv_80200"] = next_int_len(4)  # 80200
    save_stats["filibuster_stage_id"] = next_int_len(1)
    save_stats["filibuster_stage_enabled"] = next_int_len(1)

    save_stats["gv_80300"] = next_int_len(4)  # 80300

    save_stats["unknown_74"] = get_length_data()

    save_stats["gv_80500"] = next_int_len(4)  # 80500

    save_stats["unknown_75"] = get_length_data(2, 4)

    lengths = get_legend_quest_current()
    save_stats["legend_quest_current"] = lengths
    save_stats["legend_quest"] = get_legend_quest_progress(lengths)

    save_stats["unknown_133"] = get_length_data(4, 1, lengths["total"])
    save_stats["legend_quest_ids"] = get_length_data(4, 4, lengths["stages"])

    save_stats["unknown_76"] = get_data_after_leadership()
    save_stats["gv_80700"] = next_int_len(4)  # 80700
    if save_stats["dst"]:
        save_stats["unknown_104"] = next_int_len(1)
        save_stats["gv_100600"] = next_int_len(4)
        if save_stats["gv_100600"]["Value"] != 100600:
            skip(-5)
    else:
        save_stats["unknown_104"] = generate_empty_len(1)
        save_stats["gv_100600"] = generate_empty_len(4)
    save_stats["restart_pack"] = next_int_len(1)

    save_stats["unknown_101"] = get_data_after_after_leadership(save_stats["dst"])

    save_stats["medals"] = get_medals()

    save_stats["unknown_103"] = get_data_after_medals()

    lengths = get_gauntlet_current()
    save_stats["gauntlet_current"] = lengths
    save_stats["gauntlets"] = get_gauntlet_progress(lengths)

    save_stats["unknown_77"] = get_length_data(4, 1, lengths["total"])

    save_stats["gv_90300"] = next_int_len(4)  # 90300

    lengths = get_gauntlet_current()
    save_stats["unknown_78"] = lengths
    save_stats["unknown_79"] = get_gauntlet_progress(lengths)

    save_stats["unknown_80"] = get_length_data(4, 1, lengths["total"])

    save_stats["enigma_data"] = get_enigma_stages()
    data = get_cleared_slots()
    save_stats["cleared_slot_data"] = data[0]

    save_stats["unknown_121"] = data[1]

    lengths = get_gauntlet_current()
    save_stats["collab_gauntlets_current"] = lengths
    save_stats["collab_gauntlets"] = get_gauntlet_progress(lengths)
    save_stats["unknown_84"] = get_length_data(4, 1, lengths["total"])

    save_stats["unknown_85"] = data_after_after_gauntlets()

    save_stats["talent_orbs"] = get_talent_orbs(save_stats["game_version"])

    save_stats["unknown_86"] = get_data_after_orbs()

    save_stats["cat_shrine"] = get_cat_shrine_data()

    save_stats["unknown_130"] = next_int_len(4 * 5)

    save_stats["gv_90900"] = next_int_len(4)  # 90900

    save_stats["slot_names"] = get_slot_names(save_stats)
    save_stats["gv_91000"] = next_int_len(4)
    save_stats["legend_tickets"] = next_int_len(4)

    save_stats["unknown_87"] = get_length_data(1, 5)
    save_stats["unknown_88"] = next_int_len(2)

    save_stats["token"] = get_utf8_string()

    save_stats["unknown_89"] = next_int_len(1 * 3)
    save_stats["unknown_90"] = next_int_len(8)
    save_stats["unknown_91"] = next_int_len(8)

    save_stats["gv_100000"] = next_int_len(4)  # 100000
    save_stats = check_gv(save_stats, 100100)
    if save_stats["exit"]:
        return save_stats

    save_stats["date_int"] = next_int_len(4)

    save_stats["gv_100100"] = next_int_len(4)  # 100100
    save_stats = check_gv(save_stats, 100300)
    if save_stats["exit"]:
        return save_stats

    save_stats["unknown_93"] = get_length_data(4, 19, 6)

    save_stats["gv_100300"] = next_int_len(4)  # 100300
    save_stats = check_gv(save_stats, 100700)
    if save_stats["exit"]:
        return save_stats

    save_stats["unknown_94"] = get_data_near_end()

    save_stats["platinum_shards"] = next_int_len(4)

    save_stats["unknown_100"] = get_data_near_end_after_shards()

    save_stats["gv_100700"] = next_int_len(4)  # 100700
    save_stats = check_gv(save_stats, 100900)
    if save_stats["exit"]:
        return save_stats

    save_stats["aku"] = get_aku()

    save_stats["unknown_95"] = next_int_len(1 * 2)
    save_stats["unknown_96"] = get_data_after_aku()

    save_stats["gv_100900"] = next_int_len(4)  # 100900
    save_stats = check_gv(save_stats, 101000)
    if save_stats["exit"]:
        return save_stats

    save_stats["unknown_97"] = next_int_len(1)

    save_stats["gv_101000"] = next_int_len(4)  # 101000
    save_stats = check_gv(save_stats, 110000)
    if save_stats["exit"]:
        return save_stats

    save_stats["unknown_98"] = get_data_near_end_after_aku()

    save_stats["gv_110000"] = next_int_len(4)  # 110000
    save_stats = check_gv(save_stats, 110500)
    if save_stats["exit"]:
        return save_stats

    data = get_gauntlet_current()
    save_stats["behemoth_culling_current"] = data
    save_stats["behemoth_culling"] = get_gauntlet_progress(data)
    save_stats["unknown_124"] = get_length_data(4, 1, data["total"])

    save_stats["unknown_125"] = next_int_len(1)

    save_stats["gv_110500"] = next_int_len(4)  # 110500
    save_stats = check_gv(save_stats, 110600)
    if save_stats["exit"]:
        return save_stats

    save_stats["unknown_126"] = next_int_len(1)

    save_stats["gv_110600"] = next_int_len(4)  # 110600
    save_stats = check_gv(save_stats, 110700)
    if save_stats["exit"]:
        return save_stats

    save_stats["unknown_127"] = get_110700_data()

    if save_stats["dst"]:
        save_stats["unknown_128"] = next_int_len(1)
    else:
        save_stats["unknown_128"] = generate_empty_len(1)

    save_stats["gv_110700"] = next_int_len(4)  # 110700
    save_stats = check_gv(save_stats, 110800)
    if save_stats["exit"]:
        return save_stats

    save_stats["shrine_dialogs"] = next_int_len(4)

    save_stats["unknown_129"] = get_110800_data()

    save_stats["dojo_3x_speed"] = next_int_len(1)

    save_stats["unknown_132"] = get_110800_data_2()

    save_stats["gv_110800"] = next_int_len(4)  # 110800
    save_stats = check_gv(save_stats, 110800)
    if save_stats["exit"]:
        return save_stats

    length = len(save_data) - address - 32
    save_stats["extra_data"] = next_int_len(length)

    save_stats = exit_parser(save_stats)

    return save_stats



"""Handler for clearing main story chapters"""
from typing import Any


from ... import helper
from . import story_level_id_selector

CHAPTERS = [
    "Empire of Cats 1",
    "Empire of Cats 2",
    "Empire of Cats 3",
    "Into the Future 1",
    "Into the Future 2",
    "Into the Future 3",
    "Cats of the Cosmos 1",
    "Cats of the Cosmos 2",
    "Cats of the Cosmos 3",
]


def clear_specific_level_ids(
    save_stats: dict[str, Any], chapter_id: int, progress: int
) -> dict[str, Any]:
    """Clear specific levels in a chapter"""
    story_chapters = save_stats["story_chapters"]
    progress = helper.clamp(progress, 0, 48)
    if progress == 0:
        story_chapters["Chapter Progress"][chapter_id] = 0
        story_chapters["Times Cleared"][chapter_id] = [0] * 51
    else:
        stage_index = progress - 1
        story_chapters["Chapter Progress"][chapter_id] = progress
        # set all levels before the one being cleared to 1
        story_chapters["Times Cleared"][chapter_id][stage_index] = 1
        for i in range(stage_index):
            story_chapters["Times Cleared"][chapter_id][i] = 1
        # set all levels after the one being cleared to 0
        for i in range(stage_index + 1, get_total_stages(save_stats, chapter_id) + 3):
            story_chapters["Times Cleared"][chapter_id][i] = 0

    save_stats["story_chapters"] = story_chapters
    return save_stats


def has_cleared_chapter(save_stats: dict[str, Any], chapter_id: int) -> bool:
    """
    Check if a chapter has been cleared

    Args:
        save_stats (dict[str, Any]): Save stats
        chapter_id (int): Chapter ID

    Returns:
        bool: True if cleared, False if not
    """
    chapter_id = format_story_id(chapter_id)

    return save_stats["story_chapters"]["Chapter Progress"][chapter_id] >= 48


def format_story_ids(ids: list[int]) -> list[int]:
    """For some reason there is a gap after EoC 3. This adds that"""

    formatted_ids: list[int] = []
    for story_id in ids:
        formatted_ids.append(format_story_id(story_id))
    return formatted_ids


def format_story_id(chapter_id: int) -> int:
    """For some reason there is a gap after EoC 3. This adds that"""

    if chapter_id > 2:
        chapter_id += 1
    return chapter_id


def clear_levels(
    story_chapters: dict[str, Any],
    treasures: list[list[int]],
    ids: list[int],
    val: int,
    chapter_progress: int,
    clear: bool,
) -> tuple[dict[str, Any], list[list[int]]]:
    """Clear levels in a chapter"""

    for chapter_id in ids:
        story_chapters["Chapter Progress"][chapter_id] = chapter_progress
        story_chapters["Times Cleared"][chapter_id] = (
            ([val] * chapter_progress) + ([0] * (48 - chapter_progress)) + ([0] * 3)
        )
        if not clear:
            treasures[chapter_id] = [0] * 49
    return story_chapters, treasures


def get_total_stages(save_stats: dict[str, Any], chapter_id: int) -> int:
    """Get the total number of stages in a chapter"""

    return len(save_stats["story_chapters"]["Times Cleared"][chapter_id]) - 3


def clear_each(save_stats: dict[str, Any]):
    """Clear stages for each chapter"""

    chapter_ids = story_level_id_selector.select_specific_chapters()

    for chapter_id in chapter_ids:
        helper.colored_text(f"Chapter: &{chapter_id+1}& : &{CHAPTERS[chapter_id]}&")
        formatted_id = format_story_id(chapter_id)
        progress = story_level_id_selector.select_level_progress(
            chapter_id, get_total_stages(save_stats, formatted_id)
        )
        save_stats = clear_specific_level_ids(save_stats, formatted_id, progress)
    helper.colored_text("Successfully set main story chapters")
    return save_stats


def clear_all(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Clear whole chapters"""

    chapter_ids = story_level_id_selector.select_specific_chapters()
    text = ""
    for chapter_id in chapter_ids:
        text += f"Chapter: &{chapter_id+1}& : &{CHAPTERS[chapter_id]}&\n"
    helper.colored_text(text.strip("\n"))
    progress = story_level_id_selector.select_level_progress(
        None, get_total_stages(save_stats, 0)
    )
    for chapter_id in chapter_ids:
        chapter_id = format_story_id(chapter_id)
        save_stats = clear_specific_level_ids(save_stats, chapter_id, progress)
    helper.colored_text("Successfully set main story chapters")
    return save_stats



"""Handler for editing character drops"""

from typing import Any

from ... import helper, user_input_handler, csv_handler, game_data_getter
from . import cat_id_selector


def set_t_ids(save_stats: dict[str, Any]) -> dict[str, Any]:
    """handler for editing treasure ids"""

    unit_drops_stats = save_stats["unit_drops"]
    data = get_data(helper.check_data_is_jp(save_stats))

    usr_t_ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter treasures ids (Look up item drop cats battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        all_ids=data["t_ids"],
    )

    unit_drops_stats = set_t_ids_val(unit_drops_stats, data, usr_t_ids)

    save_stats["unit_drops"] = unit_drops_stats
    return save_stats


def set_c_ids(save_stats: dict[str, Any]) -> dict[str, Any]:
    """handler for editing cat ids"""

    unit_drops_stats = save_stats["unit_drops"]
    data = get_data(helper.check_data_is_jp(save_stats))

    ids = cat_id_selector.select_cats(save_stats)

    usr_c_ids = helper.check_cat_ids(ids, save_stats)
    unit_drops_stats = set_c_ids_val(unit_drops_stats, data, usr_c_ids)

    save_stats["unit_drops"] = unit_drops_stats
    return save_stats


def get_character_drops(save_stats: dict[str, Any]) -> dict[str, Any]:
    """handler for getting character drops"""

    flag_t_ids = (
        user_input_handler.colored_input(
            "Do you want to select treasure ids &(1)&, or cat ids? &(2)&:"
        )
        == "1"
    )

    if flag_t_ids:
        save_stats = set_t_ids(save_stats)
    else:
        save_stats = set_c_ids(save_stats)
    print("Successfully set unit drops")

    return save_stats


def get_data(is_jp: bool) -> dict[str, Any]:
    """gets all of the cat ids and treasure ids that can be dropped"""

    file_data = game_data_getter.get_file_latest("DataLocal", "drop_chara.csv", is_jp)
    if file_data is None:
        helper.error_text("Failed to get drop_chara.csv")
        return {"t_ids": [], "c_ids": [], "indexes": []}
    character_data = helper.parse_int_list_list(
        csv_handler.parse_csv(file_data.decode("utf-8"))[1:]
    )

    treasure_ids = helper.copy_first_n(character_data, 0)
    indexes = helper.copy_first_n(character_data, 1)
    cat_ids = helper.copy_first_n(character_data, 2)

    return {"t_ids": treasure_ids, "indexes": indexes, "c_ids": cat_ids}


def set_t_ids_val(
    unit_drops_stats: list[int], data: dict[str, Any], user_t_ids: list[int]
) -> list[int]:
    """sets the treasure ids of the unit drops"""

    for t_id in user_t_ids:
        if t_id in data["t_ids"]:
            index = data["t_ids"].index(t_id)
            save_index = data["indexes"][index]
            unit_drops_stats[save_index] = 1
    return unit_drops_stats


def set_c_ids_val(
    unit_drops_stats: list[int], data: dict[str, Any], user_t_ids: list[int]
) -> list[int]:
    """sets the cat ids of the unit drops"""

    for c_id in user_t_ids:
        if c_id in data["c_ids"]:
            index = data["c_ids"].index(c_id)
            save_index = data["indexes"][index]
            unit_drops_stats[save_index] = 1
    return unit_drops_stats



from . import (
    fix_elsewhere,
    meow_medals,
    missions,
    play_time,
    trade_progress,
    unlock_enemy_guide,
    create_new_account,
    unlock_equip_menu,
    get_gold_pass,
    claim_user_rank_rewards,
    cat_shrine,
    fix_time_issues,
    scheme_item,
)



"""Helper script for usefull functions"""

import filecmp
import json
from multiprocessing import Process
import os
import shutil
import sys
import time
from typing import Any, Callable, Generator, Optional, Union
import colored  # type: ignore

from . import (
    user_input_handler,
    server_handler,
    patcher,
    serialise_save,
    parse_save,
    tracker,
    config_manager,
)

GREEN = "#008000"
RED = "#FF0000"
DARK_YELLOW = "#D7C32A"
BLACK = "#000000"
WHITE = "#FFFFFF"
CYAN = "#00FFFF"


def get_time() -> int:
    """Get current time in seconds"""

    return int(time.time())


def get_iso_time() -> str:
    """Get the current time in iso format"""

    return time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime())


def print_line_seperator(base: str, char: str = "-", length: int = 80):
    """Print a line of a char"""
    width = shutil.get_terminal_size().columns
    if width < length:
        length = width
    colored_text(char * length, base)


def get_dirs(path: str) -> list[str]:
    """Get all directories in a path"""

    if not os.path.exists(path):
        return []

    return [dir for dir in os.listdir(path) if os.path.isdir(os.path.join(path, dir))]


def delete_dir(path: str) -> None:
    """Delete a directory and all of its contents"""

    for root, dirs, files in os.walk(path, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    os.rmdir(path)


def create_dirs(path: str) -> None:
    """Create directories if they don't exist"""

    print()


def offset_list(lst: list[int], offset: int) -> list[int]:
    """Offset each value in an list by a certain amount"""

    new_list: list[int] = []
    for item in lst:
        new_list.append(item + offset)
    return new_list


def copy_first_n(lst: list[Any], number: int) -> list[Any]:
    """Get the nth item in a list of lists"""

    new_list: list[Any] = []
    for item in lst:
        new_list.append(item[number])
    return new_list


def get_file(file_name: str) -> str:
    """Get file in files folder"""

    file_path = os.path.join(get_local_files_path(), file_name)
    return file_path


def get_files_in_dir(dir_path: str) -> list[str]:
    """Get all files in a directory"""

    files: list[str] = []
    for file in os.listdir(dir_path):
        if os.path.isfile(os.path.join(dir_path, file)):
            file_path = os.path.join(dir_path, file)
            files.append(file_path)
    return files


def find_files_in_dir(dir_path: str, file_name: str) -> list[str]:
    """Find all files in a directory with a certain name"""

    files: list[str] = []
    for file in os.listdir(dir_path):
        if file_name in file:
            files.append(file)
    return files


def get_local_files_path() -> str:
    """Get the local files path"""

    dir_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "files")
    return dir_path


def read_file_string(file_path: str) -> str:
    """Reads a file and returns its contents as a string"""

    try:
        with open(file_path, "r", encoding="utf-8") as file:
            return file.read()
    except FileNotFoundError as err:
        raise Exception("File not found: " + file_path) from err
    except UnicodeDecodeError as err:
        raise Exception("Error reading file: " + file_path + ": " + str(err)) from err


def chunks(lst: list[Any], chunk_len: int) -> Generator[Any, Any, Any]:
    """Split list into chunks of n"""

    for i in range(0, len(lst), chunk_len):
        yield lst[i : i + chunk_len]


def frames_to_time(frames: int) -> dict[str, Any]:
    """Turn frames into hours, minutes, seconds, frames"""

    frames = clamp_int(frames)
    seconds, frames = divmod(frames, 30)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return {"hh": hours, "mm": minutes, "ss": seconds, "frames": frames}


def clamp_int(value: int) -> int:
    """
    Clamp an integer to the range of a signed 32 bit integer

    Args:
        value (int): The value to clamp

    Returns:
        int: The clamped value
    """
    return clamp(value, 0, (2**31) - 1)


def num_to_bytes(num: int, length: int) -> bytes:
    """Turn number into little endian bytes"""

    return num.to_bytes(length, byteorder="little")


def seconds_to_time(seconds: int) -> dict[str, Any]:
    """Turn seconds into hours, minutes, seconds"""

    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return {"hh": hours, "mm": minutes, "ss": seconds}


def time_to_seconds(time: dict[str, Any]) -> int:
    """Turn hours, minutes, seconds into seconds"""
    seconds = time["ss"]
    seconds += time["mm"] * 60
    seconds += time["hh"] * 60 * 60
    return seconds


def time_to_frames(time: dict[str, Any]) -> int:
    """Turn hours, minutes, seconds, frames into frames"""

    total_frames = time["frames"]
    total_frames += time["ss"] * 30
    total_frames += time["mm"] * 60 * 30
    total_frames += time["hh"] * 60 * 60 * 30
    total_frames = clamp_int(total_frames)
    return total_frames


def check_int(value: str) -> Union[int, None]:
    """Check if a string is an integer"""

    value = str(value).strip(" ")
    try:
        return int(value)
    except ValueError:
        return None


def check_int_max(value: str, max_value: Optional[int] = None) -> Optional[int]:
    val = check_int(value)
    if val is None:
        return None
    if max_value is not None:
        return clamp(val, 0, max_value)
    return clamp_int(val)


def int_to_str_ls(int_list: list[int]) -> list[str]:
    """Turn list of ints to list of strings"""

    str_list: list[str] = []
    for i in int_list:
        str_list.append(str(i))
    return str_list


def parse_int_list(lst: list[str], offset: int) -> list[int]:
    """Turn string list to int list"""

    new_list: list[int] = []
    for item in lst:
        try:
            new_list.append(int(item) + offset)
        except ValueError:
            pass
    return new_list


def clamp(value: int, min_value: int, max_value: int) -> int:
    """Clamp a value between two values"""

    return max(min(value, max_value), min_value)


def write_file_bytes(file_path: str, data: bytes) -> bytes:
    """Write file as bytes"""

    try:
        with open(file_path, "wb") as file:
            file.write(data)
    except PermissionError as err:
        raise Exception("Permission denied: " + file_path) from err
    return data


def get_save_path() -> str:
    """Get the save path from the env variable"""

    save_path = os.environ.get("BC_SAVE_PATH")
    if save_path is None:
        raise Exception("BC_SAVE_PATH not set")
    return save_path


def set_save_path(path: str) -> None:
    """Set the save path in the env variable"""

    os.environ["BC_SAVE_PATH"] = os.path.abspath(path)


def get_text_splitter(isjp: bool):
    """Get the text splitter for the current save stats"""

    if isjp:
        return ","
    return "|"


def get_save_file_filetype() -> list[tuple[str, str]]:
    """Get the file types for the save file"""

    return [("Battle Cats Save Files", "*SAVE_DATA*"), ("All Files", "*.*")]


def read_file_bytes(file_path: str) -> bytes:
    """Read file as bytes"""

    with open(file_path, "rb") as file:
        return file.read()


def write_file_string(file_path: str, data: str):
    """Write file as string"""

    print()


def check_clamp(
    values: Any, max_value: int, min_value: int = 0, offset: int = -1
) -> list[int]:
    """turn a list of strings into a list of ints and clamp them between a min and max"""

    if isinstance(values, str):
        values = [values]
    int_values: list[int] = []
    for value in values:
        value = str(value).strip(" ")
        value = check_int(value)
        if value is None:
            continue
        value = clamp(value, min_value, max_value)
        value += offset
        int_values.append(value)
    return int_values


def encode_ls(lst: list[int]) -> dict[int, Any]:
    """Encode a list of integers into a dictionary"""
    return {i: lst[i] for i in range(len(lst))}


def parse_int_list_list(list_of_lists: list[list[str]]) -> list[list[Any]]:
    """Turn list of list of strings into list of list of ints"""
    new_lists: list[list[Any]] = []

    for lst in list_of_lists:
        new_list: list[Any] = []
        for item in lst:
            try:
                new_list.append(int(item))
            except ValueError:
                new_list.append(item)
        new_lists.append(new_list)
    new_lists = [line for line in new_lists if line != []]
    return new_lists


def check_hex(value: str) -> Union[str, None]:
    """Check if a string is a hex number"""

    value = str(value).strip(" ")
    try:
        int(value, 16)
        return value
    except ValueError:
        return None


def check_dec(value: str) -> Union[str, None]:
    """Check if a string is a decimal number"""

    value = str(value).strip(" ")
    try:
        int(value)
        return value
    except ValueError:
        return None


def str_to_gv(game_version: str) -> str:
    """Turn a game version with semantic versioning to integer representation"""

    split_gv = game_version.split(".")
    if len(split_gv) == 2:
        split_gv.append("0")
    final = ""
    for split in split_gv:
        final += split.zfill(2)

    return final.lstrip("0")


def gv_to_str(game_version: int) -> str:
    """Turn a game version with integer representation to semantic versioning"""

    split_gv = str(game_version).zfill(6)
    split_gv = [str(int(split_gv[i : i + 2])) for i in range(0, len(split_gv), 2)]
    return ".".join(split_gv)


def load_json(json_path: str) -> Any:
    """Load a json file"""

    return json.loads(read_file_string(json_path))


def is_jp(save_stats: dict[str, Any]) -> bool:
    """Check if the save is a Japanese save"""

    return save_stats["version"] == "jp"


def check_data_is_jp(save_stats: dict[str, Any]) -> bool:
    """Check if the save data is a Japanese save, checking the config file"""

    if config_manager.get_config_value_category("EDITOR", "ONLY_GET_EN_DATA"):
        return False
    return is_jp(save_stats)


def check_tracker(save_stats: dict[str, Any], path: str) -> None:
    """Check if the tracker is enabled"""
    item_tracker = tracker.Tracker()

    if item_tracker.has_data():
        upload = (
            user_input_handler.colored_input(
                "You have untracked bannable items that need to be uploaded. Do you want to upload them now? (&y&/&n&) (You will be unable to do so after exiting unless you have WIPE_TRACKED_ITEMS_ON_START set to False):"
            )
            == "y"
        )
        if upload:
            server_handler.meta_data_upload_handler(save_stats, path)
            colored_text("Uploaded meta data", new=GREEN)


def exit_editor():
    """Exit the editor"""

    sys.exit(0)


def check_changes(_: Any):
    """
    Check if the user wants to exit the editor

    Args:
        _ (Any): Unused
    """
    try:
        save_path = get_save_path()
    except Exception:
        return
    temp_file_path = os.path.join(
        config_manager.get_app_data_folder(), "SAVE_DATA_temp"
    )
    if not os.path.exists(temp_file_path):
        return
    is_identical = are_identical_files(save_path, temp_file_path)
    if is_identical:
        return

    ask_save_changes()


def exit_check_changes(_: Any = None):
    check_changes(None)
    exit_editor()


def ask_save_changes():
    """
    Ask if the user wants to save the changes
    """
    save = (
        user_input_handler.colored_input(
            "You have unsaved changes. Would you like to save them? (&y&/&n&):"
        )
        == "y"
    )
    if save:
        current_path = get_save_path()
        temp_file_path = os.path.join(
            config_manager.get_app_data_folder(), "SAVE_DATA_temp"
        )
        if os.path.exists(temp_file_path):
            data = read_file_bytes(temp_file_path)
            save_stats = parse_save.start_parse(data, get_country_code(data))
            check_tracker(save_stats, temp_file_path)
            write_file_bytes(current_path, read_file_bytes(temp_file_path))
            colored_text(
                f"Save data saved to &{current_path}&",
                base=GREEN,
                new=WHITE,
            )


def are_identical_files(file1: str, file2: str) -> bool:
    """Check if two files are identical"""

    return filecmp.cmp(file1, file2)


def check_cat_ids(cat_ids: list[int], save_stats: dict[str, Any]) -> list[int]:
    """Check if a list of cat ids is valid"""
    new_cat_ids: list[int] = []
    for cat_id in cat_ids:
        if cat_id > len(save_stats["cats"]) - 1:
            colored_text(f"Invalid cat id {cat_id}", base=RED)
            continue
        new_cat_ids.append(cat_id)
    return new_cat_ids


def error_text(text: str):
    """Print error text"""

    colored_text(text, base=RED)


def is_android() -> bool:
    """Check if the user is on android"""
    return "ANDROID_ROOT" in os.environ


def colored_text(
    text: str,
    base: str = WHITE,
    new: str = DARK_YELLOW,
    split_char: str = "&",
    end: str = "\n",
):
    """Print text with colors"""
    color_new = colored.fg(new)  # type: ignore
    color_base = colored.fg(base)  # type: ignore
    color_reset = colored.fg(WHITE)  # type: ignore

    text_split: list[str] = split_text(text, split_char)
    for i, text_section in enumerate(text_split):
        if i % 2:
            print(f"{color_new}{text_section}{color_base}", end="")
        else:
            print(f"{color_base}{text_section}{color_base}", end="")
    print(color_reset, end=end)


def split_text(text: str, split_char: str = "&") -> list[str]:
    """Split text on split_char, allowing for escaped split chars"""

    text_split: list[str] = []
    current_string = ""
    skip = 0
    for i, char in enumerate(text):
        if skip > 0:
            skip -= 1
            continue
        if char == "\\":
            if text[i + 1] == split_char:
                current_string += split_char
                skip = 1
                continue
        if char == split_char:
            text_split.append(current_string)
            current_string = ""
        else:
            current_string += char
    text_split.append(current_string)
    return text_split


def colored_list(
    items: list[str],
    extra_data: Any = None,
    index: bool = True,
    offset: Union[None, int] = None,
):
    """Print a list with colors and extra data if provided"""

    final = ""
    for i, item in enumerate(items):
        if index:
            final += f"{i+1}. "
        final += f"&{item}&"
        if extra_data:
            if extra_data[i] is not None:
                if isinstance(offset, int) and isinstance(extra_data[i], int):
                    final += f" &:& {extra_data[i]+offset}"
                else:
                    final += f" &:& {extra_data[i]}"
        final += "\n"
    final = final.rstrip("\n")
    colored_text(final)


def calculate_user_rank(save_stats: dict[str, Any]):
    """Calculate the user rank"""

    user_rank = 0
    for cat_id, cat_flag in enumerate(save_stats["cats"]):
        if cat_flag == 0:
            continue

        user_rank += save_stats["cat_upgrades"]["Base"][cat_id] + 1
        user_rank += save_stats["cat_upgrades"]["Plus"][cat_id]

    for skill_id in range(len(save_stats["blue_upgrades"]["Base"])):
        if skill_id == 1:
            continue
        user_rank += save_stats["blue_upgrades"]["Base"][skill_id] + 1
        user_rank += save_stats["blue_upgrades"]["Plus"][skill_id]

    return user_rank


def write_save_data(save_data: bytes, country_code: str, path: str, prompt: bool):
    """Patch the save data and write it"""

    save_data = patcher.patch_save_data(save_data, country_code)
    if prompt:
        new_path = save_file("Save File", get_save_file_filetype(), path)
        if new_path is None:
            colored_text("Save cancelled", new=RED)
            return
        path = new_path
    write_file_bytes(path, save_data)
    colored_text(f"Saved to: &{os.path.abspath(path)}&", new=GREEN)
    return save_data


def select_dir(title: str, default_dir: str) -> str:
    """
    Select a directory from the user

    Args:
        title (str): Title of the dialog
        default_dir (str): Default directory to select

    Returns:
        str: Selected directory
    """
    if not has_tkinter():
        return default_dir
    from tkinter import filedialog  # type: ignore

    root = setup_tk()
    dir_path = filedialog.askdirectory(title=title, initialdir=default_dir, parent=root)
    return dir_path


def has_tkinter() -> bool:
    """Check if tkinter is installed"""
    try:
        import tkinter  # type: ignore
    except ImportError:
        colored_text("tkinter is not installed", new=RED)
        return False
    try:
        setup_tk()
    except tkinter.TclError:
        colored_text("error setting up tkinter", new=RED)
        return False
    return True


def setup_tk() -> Any:
    """
    Setup the tkinter window

    Returns:
        Tk: Tkinter window
    """
    from tkinter import Tk

    root = Tk()
    root.withdraw()
    root.wm_attributes("-topmost", 1)  # type: ignore
    return root


def run_in_parallel(fns: list[Process]) -> None:
    """
    Run a list of functions in parallel

    Args:
        fns (list[Process]): List of functions to run in parallel
    """
    proc: list[Process] = []
    for fn in fns:
        fn.start()
        proc.append(fn)
    for p in proc:
        p.join()


def run_in_background(func: Callable[..., Any]) -> None:
    """
    Run a function in the background

    Args:
        fn (Callable[..., Any]): Function to run in the background
    """
    Process(target=func).start()


def get_cc(save_stats: dict[str, Any]) -> str:
    """Get the country code"""

    if is_jp(save_stats):
        return "jp"
    return "en"


def get_lang(jp: bool) -> str:
    """Get the language code"""

    if jp:
        return "ja"
    return "en"


def get_save_path_home() -> str:
    """
    Get the save path

    Returns:
        str: Save path
    """
    save_name = get_default_save_name()
    if config_manager.get_config_value("FIXED_SAVE_PATH"):
        path = os.path.join(get_home_path(), "bc_saves", os.path.basename(save_name))
        create_dirs(os.path.dirname(path))
        return path
    return save_name


def save_file(
    title: str, file_types: list[tuple[str, str]], path: str
) -> Optional[str]:
    """Save a file with tkinter"""
    if not has_tkinter():
        return path
    setup_tk()
    from tkinter import filedialog

    try:
        path_d = filedialog.asksaveasfile(
            mode="w",
            confirmoverwrite=True,
            initialfile=os.path.basename(path),
            filetypes=file_types,
            title=title,
        )
        if not path_d:
            return None
    except PermissionError:
        print(
            colored_text(
                "Permission denied. Make sure the file is not in use", base=RED
            )
        )
        exit_editor()
    return path_d.name


def select_file(
    title: str,
    file_types: list[tuple[str, str]],
    default_dir: str = "",
    initial_file: str = "",
) -> str:
    """Select a file with tkinter"""
    if not has_tkinter():
        path = user_input_handler.colored_input(
            f"Enter the path to the file ({title}): "
        )
        if not os.path.isfile(path):
            colored_text("Invalid path", new=RED)
            return ""
        return path
    setup_tk()
    from tkinter import filedialog

    file_path = filedialog.askopenfilename(
        initialdir=default_dir,
        title=title,
        filetypes=file_types,
        initialfile=initial_file,
    )
    return file_path


def get_home_path() -> str:
    """
    Get the home path

    Returns:
        str: Home path
    """
    return os.path.expanduser("~")


def get_default_save_name() -> str:
    """
    Get the default save name

    Returns:
        str: Default save name
    """

    save_name = config_manager.get_config_value("DEFAULT_SAVE_FILE_PATH")
    if not save_name:
        save_name = "SAVE_DATA"
    return save_name


def load_save_file(path: str) -> dict[str, Any]:
    """Load a save file, get the country code, create a backup and parse the save data"""

    save_data = read_file_bytes(path)
    country_code = get_country_code(save_data)
    colored_text(f"게임 버전(국가): &{country_code}&")
    save_stats = parse_save.start_parse(save_data, country_code)
    if config_manager.get_config_value_category("START_UP", "CREATE_BACKUP"):
        write_file_bytes(path + "_backup", save_data)
        colored_text(
            f"백업 세이브 파일: &{os.path.abspath(path)}&", new=GREEN
        )
    return {
        "save_data": save_data,
        "country_code": country_code,
        "save_stats": save_stats,
    }


def get_country_code(save_data: bytes) -> str:
    """Ask the user for their country code if it cannot be detected"""

    country_code = patcher.detect_game_version(save_data)
    if country_code is None:
        country_code = ask_cc()
    return country_code


def ask_cc():
    """Ask the user for their country code"""
    default_gv = config_manager.get_config_value("DEFAULT_COUNTRY_CODE")
    if default_gv:
        if len(default_gv) == 2:
            colored_text(f"Using default country code: &{default_gv}&")
            return default_gv

    country_code = user_input_handler.colored_input(
        "국가 선택 (&en&, &jp&, &kr&, &tw&):"
    )
    return country_code


def export_json(save_stats: dict[str, Any], path: str) -> None:
    """Export the save stats to a json file"""

    ordered_data = parse_save.re_order(save_stats)
    if os.path.isdir(path):
        path = os.path.join(path, f"{get_save_path_home()}.json")
    write_file_string(path, json.dumps(ordered_data, indent=4))
    colored_text(f"Successfully wrote json to &{os.path.abspath(path)}&")


def load_json_handler(json_path: str) -> Union[None, str]:
    """Load a save_data json file and serialise it"""

    save_stats = load_json(json_path)
    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    path = save_file(
        "Save file",
        get_save_file_filetype(),
        os.path.join(os.path.dirname(json_path), get_save_path_home()),
    )
    if path is None:
        return None
    write_file_bytes(path, save_data)
    return path



"""Handler for editting trade progress to allow for unbannable rare tickets"""
from typing import Any

from ... import helper, item


def set_trade_progress_val(storage: dict[str, Any]) -> tuple[dict[str, Any], bool]:
    """Handler for editting trade progress to allow for unbannable rare tickets"""

    space = False
    for i in range(len(storage["types"])):
        storage_item = storage["types"][i]
        if storage_item == 0 or (storage["ids"][i] == 1 and storage_item == 2):
            storage["ids"][i] = 1
            storage["types"][i] = 2
            space = True
            break
    return storage, space


def set_trade_progress(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting trade progress to allow for unbannable rare tickets"""

    trade_progress = save_stats["trade_progress"]
    max_value = helper.clamp(299 - save_stats["rare_tickets"]["Value"], 0, 299)
    storage = save_stats["cat_storage"]
    tickets = item.IntItem(
        name="Rare Tickets",
        max_value=max_value,
        value=item.Int(save_stats["rare_tickets"]["Value"]),
    )
    tickets.edit()
    trade_progress["Value"] = tickets.get_value() * 5

    storage, has_space = set_trade_progress_val(storage)

    if not has_space:
        helper.colored_text("Your cat storage is full, please free 1 space!")
        return save_stats

    save_stats["cat_storage"] = storage
    save_stats["trade_progress"] = trade_progress
    helper.colored_text(
        'You now need to go into your storage and press &"Use all"& and then press &"Trade for Ticket"&'
    )

    return save_stats



"""Update, edit and parse item tracker"""

import json
import os
from typing import Any

from . import config_manager, helper, managed_item


class TrackerItem:
    """Tracker item"""

    def __init__(
        self, value: int, managed_item_type: managed_item.ManagedItemType
    ) -> None:
        """Initialize the tracker item"""

        self.value = value
        self.managed_item_type = managed_item_type


class TrackerItems:
    """Tracker items"""

    def __init__(self, tracker_data: dict[str, int]) -> None:
        for item in managed_item.ManagedItemType:
            self.__dict__[item.value] = tracker_data[item.value]

    def to_dict(self):
        """Convert to dict"""

        return {
            item.value: self.__dict__[item.value]
            for item in managed_item.ManagedItemType
        }


class Tracker:
    """ItemTracker"""

    def __init__(self):
        self.items = TrackerItems(self.read_tracker())



    def read_tracker(self) -> dict[str, int]:
        return {
            "catfood": 0,
            "rareTicket": 0,
            "platinumTicket": 0,
            "legendTicket": 0,
        }

    def write_tracker(self):
        """Write the item tracker"""

        data = json.dumps(self.items.to_dict(), indent=4)

    def update_tracker(self, amount: int, item_type: managed_item.ManagedItemType):
        """Update the item tracker"""
        self.items.__dict__[item_type.value] += amount

        self.write_tracker()

    def set_tracker_current(self, save_stats: dict[str, Any]):
        """Set the current item tracker"""

        data = {
            "catfood": save_stats["cat_food"]["Value"],
            "rareTicket": save_stats["rare_tickets"]["Value"],
            "platinumTicket": save_stats["platinum_tickets"]["Value"],
            "legendTicket": save_stats["legend_tickets"]["Value"],
        }

        self.items = TrackerItems(data)
        self.write_tracker()

    def reset_tracker(self):
        """Reset the item tracker"""
        data = {
            "catfood": 0,
            "rareTicket": 0,
            "platinumTicket": 0,
            "legendTicket": 0,
        }

        self.items = TrackerItems(data)
        self.write_tracker()

    def has_data(self) -> bool:
        """Check if any of the items in the tracker aren't 0"""

        if not config_manager.get_config_value_category("SERVER", "UPLOAD_METADATA"):
            return False

        for item in self.items.to_dict().values():
            if item != 0:
                return True
        return False

    def parse_tracker_managed(self) -> list[managed_item.ManagedItem]:
        """Parse the item tracker as managed items"""
        data = self.items.to_dict()
        items: list[managed_item.ManagedItem] = []
        for key, value in data.items():
            if value > 0:
                detail_type = managed_item.DetailType.GET
            elif value < 0:
                detail_type = managed_item.DetailType.USE
                value = abs(value)
            else:
                continue
            items.append(
                managed_item.ManagedItem(
                    value, detail_type, managed_item.ManagedItemType(key)
                )
            )
        return items



"""Handler for editting play time"""
from typing import Any

from ... import helper, user_input_handler


def edit_play_time(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting play time"""
    play_time = save_stats["play_time"]

    hours = play_time["hh"]
    minutes = play_time["mm"]

    helper.colored_text(
        f"You currently have a play time of: &{hours}& hours and &{minutes}& minutes"
    )
    hours = helper.check_int_max(
        user_input_handler.colored_input("How many hours do you want to set?:")
    )
    minutes = helper.check_int_max(
        user_input_handler.colored_input("How many minutes do you want to set?:")
    )
    if hours is None or minutes is None or hours < 0 or minutes < 0:
        print("Please enter valid numbers")
        return save_stats
    play_time["hh"] = hours
    play_time["mm"] = minutes

    save_stats["play_time"] = play_time

    print("Successfully set play time")
    return save_stats



from . import (
    adb_handler,
    feature_handler,
    helper,
    item,
    server_handler,
    user_input_handler,
    edits,
    updater,
    patcher,
    managed_item,
    serialise_save,
    csv_handler,
    parse_save,
    config_manager,
    tracker,
    root_handler,
)



"""Handler for editing tower stages"""
from typing import Any

from . import event_stages
from ... import user_input_handler

def edit_tower(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing tower stages"""

    stage_data = save_stats["tower"]["progress"]
    stage_data = {
        "Value": stage_data,
        "Lengths": {"stars": stage_data["stars"], "stages": stage_data["stages"]},
    }

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter tower ids (Look up &Event Release Order battle cats& and scroll past the &events& and &gauntlets& to find &tower& ids) (You can enter &all& to get all, a range e.g &1&-&49&, or ids separate by spaces e.g &5 4 7&):"
        ),
        stage_data["Value"]["total"],
    )
    save_stats["tower"]["progress"] = event_stages.stage_handler(
        stage_data, ids, 0, False
    )["Value"]
    save_stats["tower"]["progress"]["total"] = stage_data["Value"]["total"]
    save_stats["tower"]["progress"]["stars"] = stage_data["Lengths"]["stars"]
    save_stats["tower"]["progress"]["stages"] = stage_data["Lengths"]["stages"]

    return save_stats



"""Handler for selecting story levels"""
from typing import Optional

from ... import user_input_handler, helper
from . import main_story


def select_specific_chapters() -> list[int]:
    """Select specific levels"""

    print("What chapters do you want to select?")
    ids = user_input_handler.select_not_inc(main_story.CHAPTERS, "clear")
    return ids


def get_option():
    """Get option"""

    options = [
        "Select specific levels with stage ids",
        "Select all levels up to a certain stage",
        "Select all levels",
    ]
    return user_input_handler.select_single(options)


def select_levels(
    chapter_id: Optional[int], forced_option: Optional[int] = None, total: int = 48
) -> list[int]:
    """Select levels"""

    if forced_option is None:
        choice = get_option()
    else:
        choice = forced_option
    if choice == 1:
        return select_specific_levels(chapter_id, total)
    if choice == 2:
        return select_levels_up_to(chapter_id, total)
    if choice == 3:
        return select_all(total)
    return []


def select_specific_levels(chapter_id: Optional[int], total: int) -> list[int]:
    """Select specific levels"""

    print("What levels do you want to select?")
    if chapter_id is not None:
        helper.colored_text(
            f"Chapter: &{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    ids = user_input_handler.get_range_ids(
        "Level ids (e.g &1&=korea, &2&=mongolia)", total
    )
    ids = helper.check_clamp(ids, total, 1, -1)
    return ids


def select_levels_up_to(chapter_id: Optional[int], total: int) -> list[int]:
    """Select levels up to a certain level"""

    print("What levels do you want to select?")
    if chapter_id is not None:
        helper.colored_text(
            f"Chapter: &{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    stage_id = user_input_handler.get_int(
        f"Enter the stage id that you want to clear/unclear up to (and including) (e.g &1&=korea cleared, &2&=korea &and& mongolia cleared, &{total}&=all)?:"
    )
    stage_id = helper.clamp(stage_id, 1, total)
    return list(range(0, stage_id))


def select_all(total: int) -> list[int]:
    """Select all levels"""

    return list(range(0, total))


def select_level_progress(
    chapter_id: Optional[int], total: int, examples: Optional[list[str]] = None
) -> int:
    """Select level progress"""

    if examples is None:
        examples = [
            "korea",
            "mongolia",
        ]

    print("What level do you want to clear up to and including?")
    if chapter_id is not None:
        helper.colored_text(
            f"Chapter: &{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    progress = user_input_handler.get_int(
        f"Enter the stage id that you want to clear/unclear (e.g &1&={examples[0]} cleared, &2&={examples[0]} &and& {examples[1]} cleared, &{total}&=all, &0&=unclear all)?:"
    )
    progress = helper.clamp(progress, 0, total)
    return progress



from typing import Any, Optional
from ... import user_input_handler, server_handler, helper, adb_handler
from ..levels import clear_tutorial


def select(save_stats: dict[str, Any]) -> dict[str, Any]:
    helper.check_changes(None)
    options = [
        "Download save data from the game using transfer and confirmation codes",
        "Select a save file from file",
        "Use adb to pull the save from a rooted device",
        "Load save data from json",
    ]
    index = (
        user_input_handler.select_single(
            options, title="Select an option to get save data:"
        )
        - 1
    )
    save_path = handle_index(index)
    if not save_path:
        return save_stats
    helper.set_save_path(save_path)
    data = helper.load_save_file(save_path)
    save_stats = data["save_stats"]
    if save_path.endswith(".json"):
        input(
            "Your save data seems to be in json format. Please use to import json option if you want to load json data.\nPress enter to continue...:"
        )
    if not clear_tutorial.is_tutorial_cleared(save_stats):
        save_stats = clear_tutorial.clear_tutorial(save_stats)
    return save_stats


def handle_index(index: int) -> Optional[str]:
    path = None
    if index == 0:
        print("Enter details for data transfer:")
        path = server_handler.download_handler()
    elif index == 1:
        print("Select save file:")
        path = helper.select_file(
            "Select a save file:",
            helper.get_save_file_filetype(),
            initial_file=helper.get_save_path_home(),
        )
    elif index == 2:
        print("Enter details for save pulling:")
        game_versions = adb_handler.find_game_versions()
        if not game_versions:
            game_version = helper.ask_cc()
        else:
            index = (
                user_input_handler.select_single(
                    game_versions, "Select", "Select a game version to pull from:", True
                )
                - 1
            )
            game_version = game_versions[index]
        path = adb_handler.adb_pull_save_data(game_version)
    elif index == 3:
        print("Select save data json file")
        js_path = helper.select_file(
            "Select save data json file",
            [("Json", "*.json")],
            initial_file=helper.get_save_path_home() + ".json",
        )
        if js_path:
            path = helper.load_json_handler(js_path)
    else:
        helper.colored_text("Please enter a recognised option", base=helper.RED)
        return None
    return path



"""Handler for editing the aku realm"""
from typing import Any
from . import story_level_id_selector, unlock_aku_realm
from ... import helper


def edit_aku(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Clear whole chapters"""

    save_stats = unlock_aku_realm.unlock_aku_realm(save_stats)

    aku = save_stats["aku"]["Value"]

    progress = story_level_id_selector.select_level_progress(None, total=49)
    progress = helper.clamp(progress, 0, 49)
    if progress == 0:
        aku["clear_progress"][0][0] = 0
        aku["clear_amount"][0][0] = [0] * len(aku["clear_amount"][0][0])

    else:
        stage_index = progress - 1
        aku["clear_progress"][0][0] = min(progress, 48)
        aku["clear_amount"][0][0][stage_index] = 1
        for i in range(stage_index):
            aku["clear_amount"][0][0][i] = 1
        for i in range(stage_index + 1, 49):
            aku["clear_amount"][0][0][i] = 0

    save_stats["aku"]["Value"] = aku
    helper.colored_text("Successfully set aku stages")
    return save_stats



"""Handler for user input"""

from typing import Any, Optional, Tuple, Union

from . import helper, locale_handler


def handle_all_at_once(
    ids: list[int],
    all_at_once: bool,
    data: list[int],
    names: list[Any],
    item_name: str,
    group_name: str,
    explain_text: str = "",
) -> list[int]:
    """Handle all at once option"""
    locale_manager = locale_handler.LocalManager.from_config()
    first = True
    value = None
    for item_id in ids:
        if all_at_once and first:
            value = helper.check_int(
                colored_input(
                    locale_manager.search_key("enter_item_name_explain")
                    % (item_name, explain_text)
                )
            )
            first = False
        elif not all_at_once:
            value = helper.check_int(
                colored_input(
                    locale_manager.search_key("enter_item_name_group_explain")
                    % (item_name, group_name, names[item_id], explain_text)
                )
            )
        if value is None:
            continue
        data[item_id] = value
    return data


def create_all_list(
    ids: list[str],
    max_val: int,
) -> dict[str, Any]:
    """Creates a list with an all at once option"""

    all_at_once = False
    if f"{max_val}" in ids:
        ids_s = list(range(1, max_val))
        ids = [format(x, "02d") for x in ids_s]
        all_at_once = True
    return {"ids": ids, "at_once": all_at_once}


def create_all_list_inc(ids: list[str], max_val: int) -> dict[str, Any]:
    """Creates a list with an all at once option and include all"""

    return create_all_list(ids, max_val)


def create_all_list_not_inc(ids: list[str], max_val: int) -> list[str]:
    """Creates a list with an all at once option and don't include all"""

    return create_all_list(ids, max_val)["ids"]


def get_range(
    usr_input: str,
    length: Union[int, None] = None,
    min_val: int = 0,
    all_ids: Union[list[int], None] = None,
) -> list[int]:
    """Get a range of numbers from user input"""
    locale_manager = locale_handler.LocalManager.from_config()
    ids: list[int] = []
    for item in usr_input.split(" "):
        if item.lower() == locale_manager.search_key("all_text").lower():
            if length is None and all_ids is None:
                helper.colored_text(
                    locale_manager.search_key("invalid_all"), helper.RED
                )
                return []
            if all_ids:
                return all_ids
            if length is not None:
                return list(range(min_val, length))
        if "-" in item:
            start_s, end_s = item.split("-")
            start = helper.check_int(start_s)
            end = helper.check_int(end_s)
            if start is None or end is None:
                helper.colored_text(
                    locale_manager.search_key("invalid_range_format"),
                    helper.RED,
                )
                return ids
            if start > end:
                start, end = end, start
            ids.extend(list(range(start, end + 1)))
        else:
            item_id = helper.check_int(item)
            if item_id is None:
                helper.colored_text(
                    locale_manager.search_key("invalid_int"), helper.RED
                )
                return ids
            ids.append(item_id)
    return ids


def colored_input(
    dialog: str, base: Optional[str] = None, new: Optional[str] = None
) -> str:
    """Format dialog as a colored string"""
    if base is None:
        base = helper.WHITE
    if new is None:
        new = helper.DARK_YELLOW
    helper.colored_text(dialog, end="", base=base, new=new)
    return input()


def get_range_ids(group_name: str, length: int) -> list[int]:
    """Get a range of ids from user input"""

    locale_manager = locale_handler.LocalManager.from_config()
    ids = get_range(
        colored_input(locale_manager.search_key("enter_range_text") % (group_name)),
        length,
    )
    return ids


def select_options(
    options: list[str],
    mode: Optional[str] = None,
    extra_data: Union[list[Any], None] = None,
    offset: int = 0,
) -> Tuple[list[int], bool]:
    """Select an option or multiple options from a list"""

    if len(options) == 1:
        return [0], True

    locale_manager = locale_handler.LocalManager.from_config()
    if mode is None:
        mode = locale_manager.search_key("edit_text")

    helper.colored_list(options, extra_data=extra_data, offset=offset)
    total = len(options)
    helper.colored_text(f"{total+1}. {locale_manager.search_key('select_all')}")
    ids_s = colored_input(
        locale_manager.search_key("select_list") % (mode, mode)
    ).split(" ")
    individual = True
    if str(total + 1) in ids_s:
        ids = list(range(1, total + 1))
        individual = False
        ids_s = helper.int_to_str_ls(ids)

    ids = helper.parse_int_list(ids_s, -1)
    for item_id in ids:
        if item_id < 0 or item_id > total - 1:
            helper.colored_text(
                locale_manager.search_key("invalid_range") % (total + 1),
                helper.RED,
            )
            return select_options(options, mode, extra_data, offset)
    return ids, individual


def select_inc(
    options: list[str],
    mode: Optional[str] = None,
    extra_data: Union[list[Any], None] = None,
    offset: int = 0,
) -> Tuple[list[int], bool]:
    """Select an option or multiple options from a list and include all"""

    return select_options(options, mode, extra_data, offset)


def select_not_inc(
    options: list[str],
    mode: Optional[str] = None,
    extra_data: Union[list[Any], None] = None,
    offset: int = 0,
) -> list[int]:
    """Select an option or multiple options from a list and don't include all"""
    return select_options(options, mode, extra_data, offset)[0]


def select_single(
    options: list[str],
    mode: Optional[str] = None,
    title: str = "",
    allow_text: bool = False,
) -> int:
    "Select a single option from a list"
    locale_manager = locale_handler.LocalManager.from_config()
    if not options:
        raise ValueError(locale_manager.search_key("error_no_options"))
    if len(options) == 1:
        return 1
    helper.colored_list(options)
    if not title:
        title = locale_manager.search_key("select_option_to") % (mode)
    val = colored_input(title)
    if allow_text:
        if val in options:
            return options.index(val) + 1
    val = helper.check_int(val)
    if val is None:
        helper.colored_text(locale_manager.search_key("invalid_int"), helper.RED)
        return select_single(options, mode, title, allow_text)
    if val < 1 or val > len(options):
        helper.colored_text(
            locale_manager.search_key("invalid_range") % (len(options)),
            helper.RED,
        )
        return select_single(options, mode, title, allow_text)
    return val


def get_int(dialog: str, default: Optional[int] = None) -> int:
    """Get user input as an integer and keep asking until a valid integer is entered"""

    helper.colored_text(dialog, end="")
    locale_manager = locale_handler.LocalManager.from_config()
    while True:
        try:
            val = input()
            val = val.strip(" ")
            return int(val)
        except ValueError:
            if default is not None:
                return default
            helper.colored_text(locale_manager.search_key("invalid_int"), helper.RED)


def ask_if_individual(item_name: str) -> bool:
    """Ask if the user wants to edit an individual item"""
    locale_manager = locale_handler.LocalManager.from_config()
    is_individual = (
        colored_input(
            locale_manager.search_key("ask_individual") % (item_name),
        )
        == "1"
    )
    return is_individual


def get_yes_no(dialog: str) -> bool:
    """Get user input as a yes or no"""
    locale_manager = locale_handler.LocalManager.from_config()
    while True:
        val = colored_input(dialog)
        if val:
            if val.lower()[0] == "y":
                return True
            if val.lower()[0] == "n":
                return False
        helper.colored_text(locale_manager.search_key("invalid_yes_no"), helper.RED)



"""Handler for editing enigma stages"""
import time
from typing import Any

from ... import helper, user_input_handler


def edit_enigma_stages(save_stats: dict[str, Any]) -> dict[str, Any]:
    """
    Edit enigma stages

    Args:
        save_stats (dict[str, Any]): Save stats

    Returns:
        dict[str, Any]: save stats
    """
    enigma_stages = save_stats["enigma_data"]

    if helper.check_data_is_jp(save_stats):
        file_name = "enigma_names_jp.txt"
    else:
        file_name = "enigma_names_en.txt"
    enigma_names = helper.read_file_string(helper.get_file(file_name)).splitlines()
    ids = user_input_handler.select_not_inc(enigma_names, "select")
    level = 3

    base_level = 25000
    for enigma_id in ids:
        abs_id = enigma_id + base_level
        data: dict[str, int] = {}
        data["level"] = level
        data["stage_id"] = abs_id
        data["decoding_status"] = 2
        data["start_time"] = int(time.time())
        enigma_stages["stages"].append(data)

    save_stats["enigma_data"] = enigma_stages

    print("Successfully edited enigma stages")
    return save_stats



"""Module that runs when the module is run directly"""
import os
import sys
import traceback

from . import (
    adb_handler,
    config_manager,
    feature_handler,
    game_data_getter,
    helper,
    parse_save,
    patcher,
    serialise_save,
    server_handler,
    tracker,
    updater,
    user_input_handler,
    root_handler,
    locale_handler,
)
from .edits.levels import clear_tutorial


def print_start_up():
    """Print start up message"""

    locale_manager = locale_handler.LocalManager.from_config()

   
    local_version = updater.get_local_version()

    print()
    if "b" in local_version:
        helper.colored_text(
            locale_manager.search_key("beta_message"),
            base=helper.RED,
            new=helper.WHITE,
        )
    print()
  


def check_update() -> None:
    """Check if there is an update available and if so, ask the user if they want to update"""
    version_info = updater.get_version_info()
    locale_manager = locale_handler.LocalManager.from_config()
    if version_info is None:
        helper.colored_text(
            locale_manager.search_key("update_check_failed"), base=helper.RED
        )
        return
    stable_ver, pre_release_ver = version_info
    


    if pre_release_ver > stable_ver:
        helper.colored_text(
            f" &|& {locale_manager.search_key('latest_pre_release_version') % pre_release_ver}",
            base=helper.CYAN,
            new=helper.WHITE,
            end="",
        )


def main():
    """Main function"""

    item_tracker = tracker.Tracker()

    if config_manager.get_config_value_category(
        "SERVER", "WIPE_TRACKED_ITEMS_ON_START"
    ):
        item_tracker.reset_tracker()
    game_data_getter.check_remove_handler()

    check_updates = config_manager.get_config_value_category(
        "START_UP", "CHECK_FOR_UPDATES"
    )
    show_start = not config_manager.get_config_value_category(
        "START_UP", "HIDE_START_TEXT"
    )
    if check_updates:
        check_update()
    if show_start:
        print_start_up()
    if show_start or check_updates:
        normal_start_up()


def normal_start_up(default_op: bool = True) -> None:
    """Display and handle options for downloading save data, pulling save data, selecting save data"""

    default_start_option = config_manager.get_config_value_category(
        "START_UP", "DEFAULT_START_OPTION"
    )
    locale_manager = locale_handler.LocalManager.from_config()
    if default_start_option != -1 and default_op:
        index = default_start_option - 1
    else:
        print()
        if not default_op:
            helper.print_line_seperator(helper.WHITE)
        options = [
            locale_manager.search_key("download_save"),
            locale_manager.search_key("select_save_file"),
            locale_manager.search_key("adb_pull_save"),
            locale_manager.search_key("load_save_data_json"),
        ]
        if helper.is_android():
            options[2] = locale_manager.search_key("android_direct_pull")
        index = (
            user_input_handler.select_single(
                options, title=locale_manager.search_key("select_save_option_title")
            )
            - 1
        )

    path = None
    if index == 0:
        helper.colored_text(locale_manager.search_key("data_transfer_message_enter"))
        path = server_handler.download_handler()
    elif index == 1:
        helper.colored_text(locale_manager.search_key("select_save_file_message"))
        path = helper.select_file(
            locale_manager.search_key("select_save_file_message"),
            helper.get_save_file_filetype(),
            initial_file=helper.get_save_path_home(),
        )
    elif index == 2:
        if not helper.is_android():
            helper.colored_text(locale_manager.search_key("adb_pull_message_enter"))
            game_versions = adb_handler.find_game_versions()
            if not game_versions:
                game_version = helper.ask_cc()
            else:
                index = (
                    user_input_handler.select_single(
                        game_versions,
                        locale_manager.search_key("select_l"),
                        locale_manager.search_key("pull_game_version_select"),
                        True,
                    )
                    - 1
                )
                game_version = game_versions[index]
            path = adb_handler.adb_pull_save_data(game_version)
        else:
            game_versions = root_handler.get_installed_battlecats_versions()
            if game_versions is not None:
                index = (
                    user_input_handler.select_single(
                        game_versions,
                        locale_manager.search_key("select_l"),
                        locale_manager.search_key("pull_game_version_select"),
                        True,
                    )
                    - 1
                )
                game_version = game_versions[index]
                path = root_handler.pull_save_data(game_version)
    elif index == 3:
        helper.colored_text(locale_manager.search_key("json_save_data_json_message"))
        js_path = helper.select_file(
            locale_manager.search_key("json_save_data_json_message"),
            [("Json", "*.json")],
            initial_file=helper.get_save_path_home() + ".json",
        )
        if js_path:
            path = helper.load_json_handler(js_path)
    else:
        helper.colored_text(locale_manager.search_key("error_option"), base=helper.RED)
        return normal_start_up(False)
    if not path:
        return normal_start_up(False)
    start(path)
    return None
def qustn1():
        locale_manager = locale_handler.LocalManager.from_config()

        helper.colored_text(locale_manager.search_key("data_transfer_message_enter"))
        path = server_handler.download_handler()
        path

def start(path: str) -> None:
    """Parse, patch, start the editor and serialise the save data"""

    locale_manager = locale_handler.LocalManager.from_config()

    if path.endswith(".json"):
        user_input_handler.colored_input(
            f"{locale_manager.search_key('error_save_json')}\n&{locale_manager.search_key('press_enter')}",
            base=helper.RED,
            new=helper.WHITE,
        )
    data = helper.load_save_file(path)
    save_stats = data["save_stats"]
    save_data: bytes = data["save_data"]
    if not clear_tutorial.is_tutorial_cleared(save_stats):
        save_stats = clear_tutorial.clear_tutorial(save_stats)
        save_data = serialise_save.start_serialize(save_stats)
    while True:
        save_stats = parse_save.start_parse(save_data, save_stats["version"])
        print(save_stats)
        save_data = patcher.patch_save_data(save_data, save_stats["version"])
        save_stats = feature_handler.menu(save_stats, path)
        save_data = serialise_save.start_serialize(save_stats)
        save_data = patcher.patch_save_data(save_data, save_stats["version"])
        if config_manager.get_config_value_category(
            "SAVE_CHANGES", "SAVE_CHANGES_ON_EDIT"
        ):
            helper.write_file_bytes(path, save_data)
            helper.colored_text(
                locale_manager.search_key("save_data_saved") % path,
                base=helper.GREEN,
                new=helper.WHITE,
            )
        temp_path = os.path.join(config_manager.get_app_data_folder(), "SAVE_DATA_temp")
        helper.write_file_bytes(temp_path, save_data)
        if config_manager.get_config_value_category(
            "SAVE_CHANGES", "ALWAYS_EXPORT_JSON"
        ):
            helper.export_json(save_stats, path + ".json")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit()
    except Exception as e:  # pylint: disable=broad-except
        try:
            locale_manager = locale_handler.LocalManager.from_config()
            error_txt = locale_manager.search_key("generic_error") % e
        except Exception:  # pylint: disable=broad-except
            error_txt = "An error has occurred: %s" % e
        helper.colored_text(
            error_txt,
            base=helper.RED,
            new=helper.WHITE,
        )
        traceback.print_exc()
        helper.exit_check_changes()
